import os
import socket
import sys
import traceback

from pandacommon.pandalogger.PandaLogger import PandaLogger
from pandaserver.dataservice import DataServiceUtils
from pandaserver.taskbuffer import JobUtils

from pandajedi.jedibrokerage import AtlasBrokerUtils
from pandajedi.jediconfig import jedi_config
from pandajedi.jedicore import JediCoreUtils
from pandajedi.jedicore.MsgWrapper import MsgWrapper

from .JumboWatchDog import JumboWatchDog
from .TypicalWatchDogBase import TypicalWatchDogBase

logger = PandaLogger().getLogger(__name__.split(".")[-1])


# watchdog for ATLAS production
class AtlasProdWatchDog(TypicalWatchDogBase):
    # constructor
    def __init__(self, taskBufferIF, ddmIF):
        TypicalWatchDogBase.__init__(self, taskBufferIF, ddmIF)
        self.pid = f"{socket.getfqdn().split('.')[0]}-{os.getpid()}-dog"

    # main
    def doAction(self):
        try:
            # get logger
            tmpLog = MsgWrapper(logger)
            tmpLog.debug("start")

            # action for priority boost
            self.doActionForPriorityBoost(tmpLog)

            # action for reassign
            self.doActionForReassign(tmpLog)

            # action for throttled
            self.doActionForThrottled(tmpLog)

            # action for high prio pending
            for minPriority, timeoutVal in [
                (950, 10),
                (900, 30),
            ]:
                self.doActionForHighPrioPending(tmpLog, minPriority, timeoutVal)

            # action to set scout job data w/o scouts
            self.doActionToSetScoutJobData(tmpLog)

            # action to throttle jobs in paused tasks
            self.doActionToThrottleJobInPausedTasks(tmpLog)

            # action for jumbo
            jumbo = JumboWatchDog(self.taskBufferIF, self.ddmIF, tmpLog, "atlas", "managed")
            jumbo.run()

            # action to provoke (mark files ready) data carousel tasks to start if DDM rules of input DS are done
            self.doActionToProvokeDCTasks(tmpLog)

        except Exception:
            errtype, errvalue = sys.exc_info()[:2]
            tmpLog.error(f"failed with {errtype.__name__}:{errvalue} {traceback.format_exc()}")
        # return
        tmpLog.debug("done")
        return self.SC_SUCCEEDED

    # action for priority boost
    def doActionForPriorityBoost(self, gTmpLog):
        # get work queue mapper
        workQueueMapper = self.taskBufferIF.getWorkQueueMap()
        # get list of work queues
        workQueueList = workQueueMapper.getAlignedQueueList(self.vo, self.prodSourceLabel)
        resource_types = self.taskBufferIF.load_resource_types()
        # loop over all work queues
        for workQueue in workQueueList:
            break_loop = False  # for special workqueues we only need to iterate once
            for resource_type in resource_types:
                gTmpLog.debug(f"start workQueue={workQueue.queue_name}")
                # get tasks to be boosted
                if workQueue.is_global_share:
                    task_criteria = {"gshare": workQueue.queue_name, "resource_type": resource_type.resource_name}
                else:
                    break_loop = True
                    task_criteria = {"workQueue_ID": workQueue.queue_id}
                dataset_criteria = {"masterID": None, "type": ["input", "pseudo_input"]}
                task_param_list = ["jediTaskID", "taskPriority", "currentPriority", "parent_tid", "gshare"]
                dataset_param_list = ["nFiles", "nFilesUsed", "nFilesTobeUsed", "nFilesFinished", "nFilesFailed"]
                taskVarList = self.taskBufferIF.getTasksWithCriteria_JEDI(
                    self.vo,
                    self.prodSourceLabel,
                    ["running"],
                    taskCriteria=task_criteria,
                    datasetCriteria=dataset_criteria,
                    taskParamList=task_param_list,
                    datasetParamList=dataset_param_list,
                    taskLockColumn="throttledTime",
                    taskLockInterval=20,
                )
                boostedPrio = 900
                toBoostRatio = 0.95
                for taskParam, datasetParam in taskVarList:
                    jediTaskID = taskParam["jediTaskID"]
                    taskPriority = taskParam["taskPriority"]
                    currentPriority = taskParam["currentPriority"]
                    parent_tid = taskParam["parent_tid"]
                    gshare = taskParam["gshare"]
                    # check parent
                    parentState = None
                    if parent_tid not in [None, jediTaskID]:
                        parentState = self.taskBufferIF.checkParentTask_JEDI(parent_tid)
                        if parentState != "completed":
                            gTmpLog.info(
                                f"#ATM #KV label=managed jediTaskID={jediTaskID} skip prio boost since parent_id={parent_tid} has parent_status={parentState}"
                            )
                            continue
                    nFiles = datasetParam["nFiles"]
                    nFilesFinished = datasetParam["nFilesFinished"]
                    nFilesFailed = datasetParam["nFilesFailed"]
                    # get num jobs
                    nJobs = self.taskBufferIF.getNumJobsForTask_JEDI(jediTaskID)
                    nRemJobs = None
                    if nJobs is not None:
                        try:
                            nRemJobs = int(float(nFiles - nFilesFinished - nFilesFailed) * float(nJobs) / float(nFiles))
                        except Exception:
                            pass
                    tmpStr = f"jediTaskID={jediTaskID} nFiles={nFiles} nFilesFinishedFailed={nFilesFinished + nFilesFailed} "
                    tmpStr += f"nJobs={nJobs} nRemJobs={nRemJobs} parent_tid={parent_tid} parentStatus={parentState}"
                    gTmpLog.debug(tmpStr)

                    try:
                        if nRemJobs is not None and float(nFilesFinished + nFilesFailed) / float(nFiles) >= toBoostRatio and nRemJobs <= 100:
                            # skip high enough
                            if currentPriority < boostedPrio:
                                gTmpLog.info(f" >>> action=priority_boosting of jediTaskID={jediTaskID} to priority={boostedPrio} #ATM #KV label=managed ")
                                self.taskBufferIF.changeTaskPriorityPanda(jediTaskID, boostedPrio)

                            # skip express or non global share
                            newShare = "Express"
                            newShareType = "managed"
                            if gshare != newShare and workQueue.is_global_share and workQueue.queue_type == newShareType:
                                gTmpLog.info(
                                    " >>> action=gshare_reassignment jediTaskID={0} from gshare_old={2} to gshare_new={1} #ATM #KV label=managed".format(
                                        jediTaskID, newShare, gshare
                                    )
                                )
                                self.taskBufferIF.reassignShare([jediTaskID], newShare, True)
                            gTmpLog.info(f">>> done jediTaskID={jediTaskID}")
                    except Exception:
                        pass

                if break_loop:
                    break

    # action for reassignment
    def doActionForReassign(self, gTmpLog):
        # get DDM I/F
        ddmIF = self.ddmIF.getInterface(self.vo)
        # get site mapper
        siteMapper = self.taskBufferIF.getSiteMapper()
        # get tasks to get reassigned
        taskList = self.taskBufferIF.getTasksToReassign_JEDI(self.vo, self.prodSourceLabel)

        gTmpLog.debug(f"got {len(taskList)} tasks to reassign")
        for taskSpec in taskList:
            tmpLog = MsgWrapper(logger, f"< jediTaskID={taskSpec.jediTaskID} >")
            tmpLog.debug("start to reassign")
            # DDM backend
            ddmBackEnd = taskSpec.getDdmBackEnd()
            # get datasets
            tmpStat, datasetSpecList = self.taskBufferIF.getDatasetsWithJediTaskID_JEDI(taskSpec.jediTaskID, ["output", "log"])
            if tmpStat is not True:
                tmpLog.error("failed to get datasets")
                continue
            # update DB
            if not taskSpec.useWorldCloud():
                # update cloudtasks
                tmpStat = self.taskBufferIF.setCloudTaskByUser("jedi", taskSpec.jediTaskID, taskSpec.cloud, "assigned", True)
                if tmpStat != "SUCCEEDED":
                    tmpLog.error("failed to update CloudTasks")
                    continue
                # check cloud
                if not siteMapper.checkCloud(taskSpec.cloud):
                    tmpLog.error(f"cloud={taskSpec.cloud} doesn't exist")
                    continue
            else:
                # re-run task brokerage
                if taskSpec.nucleus in [None, ""]:
                    taskSpec.status = "assigning"
                    taskSpec.oldStatus = None
                    taskSpec.setToRegisterDatasets()
                    self.taskBufferIF.updateTask_JEDI(taskSpec, {"jediTaskID": taskSpec.jediTaskID}, setOldModTime=True)
                    tmpLog.debug(f"#ATM #KV label=managed action=trigger_new_brokerage by setting task_status={taskSpec.status}")
                    continue

                # get nucleus
                nucleusSpec = siteMapper.getNucleus(taskSpec.nucleus)
                if nucleusSpec is None:
                    tmpLog.error(f"nucleus={taskSpec.nucleus} doesn't exist")
                    continue

                # set nucleus
                retMap = {taskSpec.jediTaskID: AtlasBrokerUtils.getDictToSetNucleus(nucleusSpec, datasetSpecList)}
                tmpRet = self.taskBufferIF.setCloudToTasks_JEDI(retMap)

            # get T1/nucleus
            if not taskSpec.useWorldCloud():
                t1SiteName = siteMapper.getCloud(taskSpec.cloud)["dest"]
            else:
                t1SiteName = nucleusSpec.getOnePandaSite()
            t1Site = siteMapper.getSite(t1SiteName)

            # loop over all datasets
            isOK = True
            for datasetSpec in datasetSpecList:
                tmpLog.debug(f"dataset={datasetSpec.datasetName}")
                if DataServiceUtils.getDistributedDestination(datasetSpec.storageToken) is not None:
                    tmpLog.debug(f"skip {datasetSpec.datasetName} is distributed")
                    continue
                # get location
                location = siteMapper.getDdmEndpoint(
                    t1Site.sitename, datasetSpec.storageToken, taskSpec.prodSourceLabel, JobUtils.translate_tasktype_to_jobtype(taskSpec.taskType)
                )
                # make subscription
                try:
                    tmpLog.debug(f"registering subscription to {location} with backend={ddmBackEnd}")
                    tmpStat = ddmIF.registerDatasetSubscription(datasetSpec.datasetName, location, "Production Output", asynchronous=True)
                    if tmpStat is not True:
                        tmpLog.error("failed to make subscription")
                        isOK = False
                        break
                except Exception:
                    errtype, errvalue = sys.exc_info()[:2]
                    tmpLog.warning(f"failed to make subscription with {errtype.__name__}:{errvalue}")
                    isOK = False
                    break
            # succeeded
            if isOK:
                # activate task
                if taskSpec.oldStatus in ["assigning", "exhausted", None]:
                    taskSpec.status = "ready"
                else:
                    taskSpec.status = taskSpec.oldStatus
                taskSpec.oldStatus = None
                self.taskBufferIF.updateTask_JEDI(taskSpec, {"jediTaskID": taskSpec.jediTaskID}, setOldModTime=True)
                tmpLog.debug("finished to reassign")

    # action for throttled tasks
    def doActionForThrottled(self, gTmpLog):
        # release tasks
        nTasks = self.taskBufferIF.releaseThrottledTasks_JEDI(self.vo, self.prodSourceLabel)
        gTmpLog.debug(f"released {nTasks} tasks")

        # throttle tasks
        nTasks = self.taskBufferIF.throttleTasks_JEDI(self.vo, self.prodSourceLabel, jedi_config.watchdog.waitForThrottled)
        gTmpLog.debug(f"throttled {nTasks} tasks")

    # action for high priority pending tasks
    def doActionForHighPrioPending(self, gTmpLog, minPriority, timeoutVal):
        timeoutForPending = None
        # try to get the timeout from the config files
        if hasattr(jedi_config.watchdog, "timeoutForPendingVoLabel"):
            timeoutForPending = JediCoreUtils.getConfigParam(jedi_config.watchdog.timeoutForPendingVoLabel, self.vo, self.prodSourceLabel)
        if timeoutForPending is None:
            timeoutForPending = jedi_config.watchdog.timeoutForPending
        timeoutForPending = int(timeoutForPending) * 24
        tmpRet = self.taskBufferIF.reactivatePendingTasks_JEDI(self.vo, self.prodSourceLabel, timeoutVal, timeoutForPending, minPriority=minPriority)
        if tmpRet is None:
            # failed
            gTmpLog.error(f"failed to reactivate high priority (>{minPriority}) tasks")
        else:
            gTmpLog.info(f"reactivated high priority (>{minPriority}) {tmpRet} tasks")

    # action to set scout job data w/o scouts
    def doActionToSetScoutJobData(self, gTmpLog):
        tmpRet = self.taskBufferIF.setScoutJobDataToTasks_JEDI(self.vo, self.prodSourceLabel)
        if tmpRet is None:
            # failed
            gTmpLog.error("failed to set scout job data")
        else:
            gTmpLog.info("set scout job data successfully")

    # action to throttle jobs in paused tasks
    def doActionToThrottleJobInPausedTasks(self, gTmpLog):
        tmpRet = self.taskBufferIF.throttleJobsInPausedTasks_JEDI(self.vo, self.prodSourceLabel)
        if tmpRet is None:
            # failed
            gTmpLog.error("failed to thottle jobs in paused tasks")
        else:
            for jediTaskID, pandaIDs in tmpRet.items():
                gTmpLog.info(f"throttled jobs in paused jediTaskID={jediTaskID} successfully")
                tmpRet = self.taskBufferIF.killJobs(pandaIDs, "reassign", "51", True)
                gTmpLog.info(f"reassigned {len(pandaIDs)} jobs in paused jediTaskID={jediTaskID} with {tmpRet}")

    # action to provoke (mark files ready) data carousel tasks to start if DDM rules of input DS are done
    def doActionToProvokeDCTasks(self, gTmpLog):
        # lock
        got_lock = self.taskBufferIF.lockProcess_JEDI(
            vo=self.vo,
            prodSourceLabel=self.prodSourceLabel,
            cloud=None,
            workqueue_id=None,
            resource_name=None,
            component="AtlasProdWatchDog.doActionToProvokeDCT",
            pid=self.pid,
            timeLimit=30,
        )
        if not got_lock:
            gTmpLog.debug("doActionToProvokeDCTasks locked by another process. Skipped")
            return
        # run
        res_dict = self.taskBufferIF.get_pending_dc_tasks_JEDI(task_type="prod", time_limit_minutes=60)
        if res_dict is None:
            # failed
            gTmpLog.error("failed to get pending DC tasks")
        elif not res_dict:
            # empty
            gTmpLog.debug("no pending DC task; skipped")
        else:
            gTmpLog.debug(f"got {len(res_dict)} DC tasks to provoke")
            ddm_if = self.ddmIF.getInterface(self.vo)
            # loop over pending DC tasks
            for task_id, ds_name_list in res_dict.items():
                if not ds_name_list:
                    continue
                total_all_ok = True
                for ds_name in ds_name_list:
                    ret_data = ddm_if.get_rules_state(ds_name)
                    try:
                        all_ok, rule_dict = ret_data
                        total_all_ok = total_all_ok and all_ok
                    except ValueError:
                        gTmpLog.error(f"failed to get rule info for task={task_id}. data={ret_data}")
                        total_all_ok = False
                        break
                if total_all_ok:
                    # all rules ok; provoke the task
                    gTmpLog.info(f"provoking task {task_id}")
                    self.taskBufferIF.updateInputDatasetsStagedAboutIdds_JEDI(task_id, None, None)
                    gTmpLog.info(f"all staging rules of task {task_id} are OK; provoked")
                else:
                    gTmpLog.debug(f"not all staging rules of task {task_id} are OK; skipped ")



import re
import sys

# logger
from pandacommon.pandalogger.PandaLogger import PandaLogger
from pandajedi.jedicore.MsgWrapper import MsgWrapper

from .TypicalWatchDogBase import TypicalWatchDogBase

logger = PandaLogger().getLogger(__name__.split(".")[-1])


# watchdog for general purpose
class GenWatchDog(TypicalWatchDogBase):
    # constructor
    def __init__(self, taskBufferIF, ddmIF):
        TypicalWatchDogBase.__init__(self, taskBufferIF, ddmIF)

    # main
    def doAction(self):
        # get logger
        tmpLog = MsgWrapper(logger)
        tmpLog.debug("start")
        # return
        tmpLog.debug("done")
        return self.SC_SUCCEEDED



"""
file specification for JEDI

"""

import re
import types

from taskbuffer.FileSpec import FileSpec as JobFileSpec


class JediFileSpec(object):
    # attributes
    _attributes = (
        "jediTaskID",
        "datasetID",
        "fileID",
        "creationDate",
        "lastAttemptTime",
        "lfn",
        "GUID",
        "type",
        "status",
        "fsize",
        "checksum",
        "scope",
        "attemptNr",
        "maxAttempt",
        "nEvents",
        "keepTrack",
        "startEvent",
        "endEvent",
        "firstEvent",
        "boundaryID",
        "PandaID",
        "failedAttempt",
        "lumiBlockNr",
        "outPandaID",
        "maxFailure",
        "ramCount",
        "is_waiting",
        "proc_status",
    )
    # attributes which have 0 by default
    _zeroAttrs = ("fsize", "attemptNr", "failedAttempt", "ramCount")
    # mapping between sequence and attr
    _seqAttrMap = {"fileID": "ATLAS_PANDA.JEDI_DATASET_CONT_FILEID_SEQ.nextval"}

    # constructor
    def __init__(self):
        # install attributes
        for attr in self._attributes:
            if attr in self._zeroAttrs:
                object.__setattr__(self, attr, 0)
            else:
                object.__setattr__(self, attr, None)
        # map of changed attributes
        object.__setattr__(self, "_changedAttrs", {})
        # locality
        object.__setattr__(self, "locality", {})
        # source name
        object.__setattr__(self, "sourceName", None)

    # override __setattr__ to collecte the changed attributes
    def __setattr__(self, name, value):
        oldVal = getattr(self, name)
        object.__setattr__(self, name, value)
        newVal = getattr(self, name)
        # collect changed attributes
        if oldVal != newVal:
            self._changedAttrs[name] = value

    # reset changed attribute list
    def resetChangedList(self):
        object.__setattr__(self, "_changedAttrs", {})

    # return map of values
    def valuesMap(self, useSeq=False, onlyChanged=False):
        ret = {}
        for attr in self._attributes:
            # use sequence
            if useSeq and attr in self._seqAttrMap:
                continue
            # only changed attributes
            if onlyChanged:
                if attr not in self._changedAttrs:
                    continue
            val = getattr(self, attr)
            if val is None:
                if attr in self._zeroAttrs:
                    val = 0
                else:
                    val = None
            ret[f":{attr}"] = val
        return ret

    # pack tuple into FileSpec
    def pack(self, values):
        for i in range(len(self._attributes)):
            attr = self._attributes[i]
            val = values[i]
            object.__setattr__(self, attr, val)

    # return column names for INSERT
    def columnNames(cls, useSeq=False, defaultVales=None, skipDefaultAttr=False):
        if defaultVales is None:
            defaultVales = {}
        ret = ""
        for attr in cls._attributes:
            if skipDefaultAttr and (attr in cls._seqAttrMap or attr in defaultVales):
                continue
            if ret != "":
                ret += ","
            if useSeq and attr in cls._seqAttrMap:
                ret += f"{cls._seqAttrMap[attr]}"
                continue
            if attr in defaultVales:
                arg = defaultVales[attr]
                if arg is None:
                    ret += "NULL"
                elif isinstance(arg, str):
                    ret += f"'{arg}'"
                else:
                    ret += f"{arg}"
                continue
            ret += attr
        return ret

    columnNames = classmethod(columnNames)

    # return expression of bind variables for INSERT
    def bindValuesExpression(cls, useSeq=True):
        ret = "VALUES("
        for attr in cls._attributes:
            if useSeq and attr in cls._seqAttrMap:
                ret += f"{cls._seqAttrMap[attr]},"
            else:
                ret += f":{attr},"
        ret = ret[:-1]
        ret += ")"
        return ret

    bindValuesExpression = classmethod(bindValuesExpression)

    # return an expression of bind variables for UPDATE to update only changed attributes
    def bindUpdateChangesExpression(self):
        ret = ""
        for attr in self._attributes:
            if attr in self._changedAttrs:
                ret += f"{attr}=:{attr},"
        ret = ret[:-1]
        ret += " "
        return ret

    # convert to job's FileSpec
    def convertToJobFileSpec(self, datasetSpec, setType=None, useEventService=False):
        jobFileSpec = JobFileSpec()
        jobFileSpec.fileID = self.fileID
        jobFileSpec.datasetID = datasetSpec.datasetID
        jobFileSpec.jediTaskID = datasetSpec.jediTaskID
        jobFileSpec.lfn = self.lfn
        jobFileSpec.GUID = self.GUID
        if setType is None:
            jobFileSpec.type = self.type
        else:
            jobFileSpec.type = setType
        jobFileSpec.scope = self.scope
        jobFileSpec.fsize = self.fsize
        jobFileSpec.checksum = self.checksum
        jobFileSpec.attemptNr = self.attemptNr
        # dataset attribute
        if datasetSpec is not None:
            # dataset
            if datasetSpec.containerName not in [None, ""]:
                jobFileSpec.dataset = datasetSpec.containerName
            else:
                jobFileSpec.dataset = datasetSpec.datasetName
            if self.type in datasetSpec.getInputTypes() or setType in datasetSpec.getInputTypes():
                # prodDBlock
                jobFileSpec.prodDBlock = datasetSpec.datasetName
                # storage token
                if datasetSpec.storageToken not in ["", None]:
                    jobFileSpec.dispatchDBlockToken = datasetSpec.storageToken
            else:
                # destinationDBlock
                jobFileSpec.destinationDBlock = datasetSpec.datasetName
                # storage token
                if datasetSpec.storageToken not in ["", None]:
                    jobFileSpec.destinationDBlockToken = datasetSpec.storageToken.split("/")[0]
                # destination
                if datasetSpec.destination not in ["", None]:
                    jobFileSpec.destinationSE = datasetSpec.destination
                # set prodDBlockToken for Event Service
                if useEventService and datasetSpec.getObjectStore() is not None:
                    jobFileSpec.prodDBlockToken = f"objectstore^{datasetSpec.getObjectStore()}"
                # allow no output
                if datasetSpec.isAllowedNoOutput():
                    jobFileSpec.allowNoOutput()
        # return
        return jobFileSpec

    # convert from job's FileSpec
    def convertFromJobFileSpec(self, jobFileSpec):
        self.fileID = jobFileSpec.fileID
        self.datasetID = jobFileSpec.datasetID
        self.jediTaskID = jobFileSpec.jediTaskID
        self.lfn = jobFileSpec.lfn
        self.GUID = jobFileSpec.GUID
        self.type = jobFileSpec.type
        self.scope = jobFileSpec.scope
        self.fsize = jobFileSpec.fsize
        self.checksum = jobFileSpec.checksum
        self.attemptNr = jobFileSpec.attemptNr
        # convert NULL to None
        for attr in self._attributes:
            val = getattr(self, attr)
            if val == "NULL":
                object.__setattr__(self, attr, None)
        # return
        return

    # check if event-level splitting is used
    def doEventLevelSplit(self):
        if self.startEvent is None:
            return False
        return True

    # get effective number of events
    def getEffectiveNumEvents(self):
        if self.endEvent is not None and self.startEvent is not None:
            evtCounts = self.endEvent - self.startEvent + 1
            if evtCounts > 0:
                return evtCounts
            return 1
        if self.nEvents is not None and self.nEvents > 0:
            return self.nEvents
        return 1

    # extract fields string
    def extractFieldsStr(self, fieldNumList):
        tmpFieldStr = ""
        try:
            tmpMidStrList = re.split("\.|_tid\d+", self.lfn)
            for tmpFieldNum in fieldNumList:
                tmpFieldStr += "." + tmpMidStrList[tmpFieldNum - 1]
        except Exception:
            pass
        return tmpFieldStr



import json
import os
import socket

from pandacommon.pandalogger import logger_utils
from pandajedi.jedicore.ThreadUtils import ListWithLock, ThreadPool
from pandajedi.jediddm.DDMInterface import DDMInterface
from pandajedi.jedimsgprocessor.base_msg_processor import BaseMsgProcPlugin
from pandajedi.jediorder.JobGenerator import JobGeneratorThread
from pandajedi.jediorder.TaskSetupper import TaskSetupper

# logger
base_logger = logger_utils.setup_logger(__name__.split(".")[-1])


# plugin to process messages from Panda to JEDI
class PandaToJediMsgProcPlugin(BaseMsgProcPlugin):
    def initialize(self):
        BaseMsgProcPlugin.initialize(self)
        self.ddmIF = DDMInterface()
        self.ddmIF.setupInterface()
        self.pid = f"{socket.getfqdn().split('.')[0]}-{os.getpid()}_{os.getpgrp()}-pjmsg"

    def process(self, msg_obj, decoded_data=None):
        # logger
        tmp_log = logger_utils.make_logger(base_logger, token=self.get_pid(), method_name="process")
        # start
        tmp_log.info("start")
        # parse
        if decoded_data is None:
            # json decode
            try:
                msg_dict = json.loads(msg_obj.data)
            except Exception as e:
                err_str = f"failed to parse message json {msg_obj.data} , skipped. {e.__class__.__name__} : {e}"
                tmp_log.error(err_str)
                raise
        else:
            msg_dict = decoded_data
        # run
        try:
            tmp_log.debug(f"got message {msg_dict}")
            if msg_dict["msg_type"] == "generate_job":
                # get task to generate jobs
                jediTaskID = int(msg_dict["taskid"])
                s, taskSpec = self.tbIF.getTaskWithID_JEDI(jediTaskID)
                if not taskSpec:
                    tmp_log.debug(f"unknown task {jediTaskID}")
                else:
                    # get WQ
                    vo = taskSpec.vo
                    prodSourceLabel = taskSpec.prodSourceLabel
                    workQueue = self.tbIF.getWorkQueueMap().getQueueWithIDGshare(taskSpec.workQueue_ID, taskSpec.gshare)
                    # get inputs
                    tmpList = self.tbIF.getTasksToBeProcessed_JEDI(self.pid, None, workQueue, None, None, nFiles=1000, target_tasks=[jediTaskID])
                    if tmpList:
                        inputList = ListWithLock(tmpList)
                        # create thread
                        threadPool = ThreadPool()
                        siteMapper = self.tbIF.getSiteMapper()
                        taskSetupper = TaskSetupper(vo, prodSourceLabel)
                        taskSetupper.initializeMods(self.tbIF, self.ddmIF)
                        gen = JobGeneratorThread(
                            inputList, threadPool, self.tbIF, self.ddmIF, siteMapper, True, taskSetupper, self.pid, workQueue, "pjmsg", None, None, None, False
                        )
                        gen.start()
                        gen.join()
            else:
                tmp_log.debug(f"unknown message type : {msg_dict['msg_type']}")
        except Exception as e:
            err_str = f"failed to run, skipped. {e.__class__.__name__} : {e}"
            tmp_log.error(err_str)
            raise
        # done
        tmp_log.info("done")



from pandajedi.jedicore.Interaction import CommandReceiveInterface


# base class to interact with DDM
class DDMClientBase(CommandReceiveInterface):
    # constructor
    def __init__(self, con):
        CommandReceiveInterface.__init__(self, con)

    # list dataset/container
    def listDatasets(self, datasetName, ignorePandaDS=True):
        return self.SC_SUCCEEDED, [datasetName]



import datetime
import multiprocessing
import multiprocessing.reduction
import os
import signal
import sys
import time

try:
    from multiprocessing.connection import reduce_connection
except ImportError:
    from multiprocessing.reduction import reduce_connection

# import multiprocessing
# logger = multiprocessing.log_to_stderr()
# logger.setLevel(multiprocessing.SUBDEBUG)

###########################################################
#
# status codes for IPC
#


# class for status code
class StatusCode(object):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return f"{self.value}"

    # comparator
    def __eq__(self, other):
        try:
            return self.value == other.value
        except Exception:
            return False

    def __ne__(self, other):
        try:
            return self.value != other.value
        except Exception:
            return True


# mapping to accessors
statusCodeMap = {
    "SC_SUCCEEDED": StatusCode(0),
    "SC_FAILED": StatusCode(1),
    "SC_FATAL": StatusCode(2),
}


# install the list of status codes to a class
def installSC(cls):
    for sc, val in statusCodeMap.items():
        setattr(cls, sc, val)


# install SCs in this module
installSC(sys.modules[__name__])


###########################################################
#
# classes for IPC
#


# log message with timestamp
def dumpStdOut(sender, message):
    timeNow = datetime.datetime.now(datetime.timezone.utc).replace(tzinfo=None)
    print(f"{str(timeNow)} {sender}: INFO    {message}")


# object class for command
class CommandObject(object):
    # constructor
    def __init__(self, methodName, argList, argMap):
        self.methodName = methodName
        self.argList = argList
        self.argMap = argMap


# object class for response
class ReturnObject(object):
    # constructor
    def __init__(self):
        self.statusCode = None
        self.errorValue = None
        self.returnValue = None


# process class
class ProcessClass(object):
    # constructor
    def __init__(self, pid, connection):
        self.pid = pid
        self.nused = 0
        self.usedMemory = 0
        self.nMemLookup = 20
        # reduce connection to make it picklable
        self.reduced_pipe = reduce_connection(connection)

    # get connection
    def connection(self):
        # rebuild connection
        return self.reduced_pipe[0](*self.reduced_pipe[1])

    # reduce connection
    def reduceConnection(self, connection):
        self.reduced_pipe = reduce_connection(connection)

    # get memory usage
    def getMemUsage(self):
        # update memory info
        if self.nused % self.nMemLookup == 0:
            try:
                # read memory info from /proc
                t = open(f"/proc/{self.pid}/status")
                v = t.read()
                t.close()
                value = 0
                for line in v.split("\n"):
                    if line.startswith("VmRSS"):
                        items = line.split()
                        value = int(items[1])
                        if items[2] in ["kB", "KB"]:
                            value /= 1024
                        elif items[2] in ["mB", "MB"]:
                            pass
                        break
                self.usedMemory = value
            except Exception:
                pass
            return self.usedMemory
        else:
            return None


# method class
class MethodClass(object):
    # constructor
    def __init__(self, className, methodName, vo, connectionQueue, voIF):
        self.className = className
        self.methodName = methodName
        self.vo = vo
        self.connectionQueue = connectionQueue
        self.voIF = voIF
        self.pipeList = []

    # method emulation
    def __call__(self, *args, **kwargs):
        commandObj = CommandObject(self.methodName, args, kwargs)
        nTry = 3
        for iTry in range(nTry):
            # exceptions
            retException = None
            strException = None
            try:
                stepIdx = 0
                # get child process
                child_process = self.connectionQueue.get()
                # get pipe
                stepIdx = 1
                pipe = child_process.connection()
                # get ack
                stepIdx = 2
                timeoutPeriodACK = 30
                if not pipe.poll(timeoutPeriodACK):
                    raise JEDITimeoutError(f"did not get ACK for {timeoutPeriodACK}sec")
                ack = pipe.recv()
                # send command
                stepIdx = 3
                pipe.send(commandObj)
                # wait response
                stepIdx = 4
                timeoutPeriod = 600
                timeNow = datetime.datetime.now(datetime.timezone.utc).replace(tzinfo=None)
                if not pipe.poll(timeoutPeriod):
                    raise JEDITimeoutError(f"did not get response for {timeoutPeriod}sec")
                regTime = datetime.datetime.now(datetime.timezone.utc).replace(tzinfo=None) - timeNow
                if regTime > datetime.timedelta(seconds=60):
                    dumpStdOut(
                        self.className,
                        f"methodName={self.methodName} took {regTime.seconds}.{int(regTime.microseconds / 1000):03d} sec in pid={child_process.pid}",
                    )
                # get response
                stepIdx = 5
                ret = pipe.recv()
                # set exception type based on error
                stepIdx = 6
                if ret.statusCode == SC_FAILED:
                    retException = JEDITemporaryError
                elif ret.statusCode == SC_FATAL:
                    retException = JEDIFatalError
            except Exception:
                errtype, errvalue = sys.exc_info()[:2]
                retException = errtype
                argStr = f"args={str(args)} kargs={str(kwargs)}"
                strException = f"VO={self.vo} type={errtype.__name__} stepIdx={stepIdx} : {self.className}.{self.methodName} {errvalue} {argStr[:200]}"
            # increment nused
            child_process.nused += 1
            # memory check
            largeMemory = False
            memUsed = child_process.getMemUsage()
            if memUsed is not None:
                memStr = f"pid={child_process.pid} memory={memUsed}MB"
                if memUsed > 1.5 * 1024:
                    largeMemory = True
                    memStr += " exceeds memory limit"
                    dumpStdOut(self.className, memStr)
            # kill old or problematic process
            if child_process.nused > 1000 or retException not in [None, JEDITemporaryError, JEDIFatalError] or largeMemory:
                dumpStdOut(
                    self.className,
                    f"methodName={self.methodName} ret={retException} nused={child_process.nused} {strException} in pid={child_process.pid}",
                )
                # close connection
                try:
                    pipe.close()
                except Exception:
                    pass
                # terminate child process
                try:
                    dumpStdOut(self.className, f"killing pid={child_process.pid}")
                    os.kill(child_process.pid, signal.SIGKILL)
                    dumpStdOut(self.className, f"waiting pid={child_process.pid}")
                    os.waitpid(child_process.pid, 0)
                    dumpStdOut(self.className, f"terminated pid={child_process.pid}")
                except Exception:
                    errtype, errvalue = sys.exc_info()[:2]
                    if "No child processes" not in str(errvalue):
                        dumpStdOut(self.className, f"failed to terminate {child_process.pid} with {errtype}:{errvalue}")
                # make new child process
                self.voIF.launchChild()
            else:
                # reduce process object to avoid deadlock due to rebuilding of connection
                child_process.reduceConnection(pipe)
                self.connectionQueue.put(child_process)
            # success, fatal error, or maximally attempted
            if retException in [None, JEDIFatalError] or (iTry + 1 == nTry):
                break
            # sleep
            time.sleep(1)
        # raise exception
        if retException is not None:
            if strException is None:
                strException = f"VO={self.vo} {ret.errorValue}"
            raise retException(strException)
        # return
        if ret.statusCode == SC_SUCCEEDED:
            return ret.returnValue
        else:
            raise retException(f"VO={self.vo} {ret.errorValue}")


# interface class to send command
class CommandSendInterface(object):
    # constructor
    def __init__(self, vo, maxChild, moduleName, className):
        self.vo = vo
        self.maxChild = maxChild
        self.connectionQueue = multiprocessing.Queue(maxChild)
        self.moduleName = moduleName
        self.className = className

    # factory method
    def __getattr__(self, attrName):
        return MethodClass(self.className, attrName, self.vo, self.connectionQueue, self)

    # launcher for child processe
    def launcher(self, channel):
        # import module
        mod = __import__(self.moduleName)
        for subModuleName in self.moduleName.split(".")[1:]:
            mod = getattr(mod, subModuleName)
        # get class
        cls = getattr(mod, self.className)
        # start child process
        msg = f"start {self.className} with pid={os.getpid()}"
        dumpStdOut(self.moduleName, msg)
        timeNow = datetime.datetime.now(datetime.timezone.utc).replace(tzinfo=None)
        try:
            cls(channel).start()
        except Exception:
            errtype, errvalue = sys.exc_info()[:2]
            dumpStdOut(self.className, f"launcher crashed with {errtype}:{errvalue}")

    # launch child processes to interact with DDM
    def launchChild(self):
        # make pipe
        parent_conn, child_conn = multiprocessing.Pipe()
        # make child process
        child_process = multiprocessing.Process(target=self.launcher, args=(child_conn,))
        # start child process
        child_process.daemon = True
        child_process.start()
        # keep process in queue
        processObj = ProcessClass(child_process.pid, parent_conn)
        # sync to wait until object in the child process is instantiated
        pipe = processObj.connection()
        pipe.recv()
        processObj.reduceConnection(pipe)
        # ready
        self.connectionQueue.put(processObj)

    # initialize
    def initialize(self):
        for i in range(self.maxChild):
            self.launchChild()


# interface class to receive command
class CommandReceiveInterface(object):
    # constructor
    def __init__(self, con):
        self.con = con
        self.cacheMap = {}

    # make key for cache
    def makeKey(self, className, methodName, argList, argMap):
        try:
            tmpKey = f"{className}:{methodName}:"
            for argItem in argList:
                tmpKey += f"{str(argItem)}:"
            for argKey, argVal in argMap.items():
                tmpKey += f"{argKey}={str(argVal)}:"
            tmpKey = tmpKey[:-1]
            return tmpKey
        except Exception:
            return None

    # main loop
    def start(self):
        # sync
        self.con.send("ready")
        # main loop
        while True:
            # send ACK
            self.con.send("ack")
            # get command
            commandObj = self.con.recv()
            # make return
            retObj = ReturnObject()
            # get class name
            className = self.__class__.__name__
            # check method name
            if not hasattr(self, commandObj.methodName):
                # method not found
                retObj.statusCode = self.SC_FATAL
                retObj.errorValue = f"type=AttributeError : {className} instance has no attribute {commandObj.methodName}"
            else:
                try:
                    # use cache
                    useCache = False
                    doExec = True
                    if "useResultCache" in commandObj.argMap:
                        # get time range
                        timeRange = commandObj.argMap["useResultCache"]
                        # delete from args map
                        del commandObj.argMap["useResultCache"]
                        # make key for cache
                        tmpCacheKey = self.makeKey(className, commandObj.methodName, commandObj.argList, commandObj.argMap)
                        if tmpCacheKey is not None:
                            useCache = True
                            # cache is fresh
                            if tmpCacheKey in self.cacheMap and self.cacheMap[tmpCacheKey]["utime"] + datetime.timedelta(
                                seconds=timeRange
                            ) > datetime.datetime.now(datetime.timezone.utc).replace(tzinfo=None):
                                tmpRet = self.cacheMap[tmpCacheKey]["value"]
                                doExec = False
                    # exec
                    if doExec:
                        # get function
                        functionObj = getattr(self, commandObj.methodName)
                        # exec
                        tmpRet = functionObj(*commandObj.argList, **commandObj.argMap)
                    if isinstance(tmpRet, StatusCode):
                        # only status code was returned
                        retObj.statusCode = tmpRet
                    elif (isinstance(tmpRet, tuple) or isinstance(tmpRet, list)) and len(tmpRet) > 0 and isinstance(tmpRet[0], StatusCode):
                        retObj.statusCode = tmpRet[0]
                        # status code + return values
                        if len(tmpRet) > 1:
                            if retObj.statusCode == self.SC_SUCCEEDED:
                                if len(tmpRet) == 2:
                                    retObj.returnValue = tmpRet[1]
                                else:
                                    retObj.returnValue = tmpRet[1:]
                            else:
                                if len(tmpRet) == 2:
                                    retObj.errorValue = tmpRet[1]
                                else:
                                    retObj.errorValue = tmpRet[1:]
                    else:
                        retObj.statusCode = self.SC_SUCCEEDED
                        retObj.returnValue = tmpRet
                except Exception:
                    errtype, errvalue = sys.exc_info()[:2]
                    # failed
                    retObj.statusCode = self.SC_FATAL
                    retObj.errorValue = f"type={errtype.__name__} : {className}.{commandObj.methodName} : {errvalue}"
                # cache
                if useCache and doExec and retObj.statusCode == self.SC_SUCCEEDED:
                    self.cacheMap[tmpCacheKey] = {"utime": datetime.datetime.now(datetime.timezone.utc).replace(tzinfo=None), "value": tmpRet}
            # return
            self.con.send(retObj)


# install SCs
installSC(CommandReceiveInterface)


###########################################################
#
# exceptions for IPC
#


# exception for temporary error
class JEDITemporaryError(Exception):
    pass


# exception for fatal error
class JEDIFatalError(Exception):
    pass


# exception for timeout error
class JEDITimeoutError(Exception):
    pass


# exception for no child error
class JEDINoChildError(Exception):
    pass



import random
import re

from pandacommon.pandalogger.PandaLogger import PandaLogger
from pandajedi.jedicore import Interaction, JediCoreUtils
from pandajedi.jedicore.MsgWrapper import MsgWrapper
from pandajedi.jedicore.SiteCandidate import SiteCandidate
from pandajedi.jedirefine import RefinerUtils

from . import AtlasBrokerUtils
from .JobBrokerBase import JobBrokerBase

logger = PandaLogger().getLogger(__name__.split(".")[-1])


# brokerage for general purpose
class GenJobBroker(JobBrokerBase):
    # constructor
    def __init__(self, ddmIF, taskBufferIF):
        JobBrokerBase.__init__(self, ddmIF, taskBufferIF)

    # main
    def doBrokerage(self, taskSpec, cloudName, inputChunk, taskParamMap):
        # make logger
        tmpLog = MsgWrapper(logger, f"<jediTaskID={taskSpec.jediTaskID}>")
        tmpLog.debug("start")
        # return for failure
        retFatal = self.SC_FATAL, inputChunk
        retTmpError = self.SC_FAILED, inputChunk
        # set cloud
        try:
            if not taskParamMap:
                taskParam = self.taskBufferIF.getTaskParamsWithID_JEDI(taskSpec.jediTaskID)
                taskParamMap = RefinerUtils.decodeJSON(taskParam)
            if not taskSpec.cloud and "cloud" in taskParamMap:
                taskSpec.cloud = taskParamMap["cloud"]
        except Exception:
            pass
        # get sites in the cloud
        site_preassigned = True
        if taskSpec.site not in ["", None]:
            tmpLog.debug(f"site={taskSpec.site} is pre-assigned")
            if self.siteMapper.checkSite(taskSpec.site):
                scanSiteList = [taskSpec.site]
            else:
                scanSiteList = []
                for tmpSite in self.siteMapper.getCloud(taskSpec.cloud)["sites"]:
                    if re.search(taskSpec.site, tmpSite):
                        scanSiteList.append(tmpSite)
                if not scanSiteList:
                    tmpLog.error(f"unknown site={taskSpec.site}")
                    taskSpec.setErrDiag(tmpLog.uploadLog(taskSpec.jediTaskID))
                    return retTmpError
        elif inputChunk.getPreassignedSite() is not None:
            scanSiteList = [inputChunk.getPreassignedSite()]
            tmpLog.debug(f"site={inputChunk.getPreassignedSite()} is pre-assigned in masterDS")
        else:
            site_preassigned = False
            scanSiteList = self.siteMapper.getCloud(taskSpec.cloud)["sites"]
            # remove NA
            if "NA" in scanSiteList:
                scanSiteList.remove("NA")
            tmpLog.debug(f"cloud={taskSpec.cloud} has {len(scanSiteList)} candidates")
        tmpLog.debug(f"initial {len(scanSiteList)} candidates")
        ######################################
        # selection for status and PandaSite
        newScanSiteList = []
        for tmpSiteName in scanSiteList:
            tmpSiteSpec = self.siteMapper.getSite(tmpSiteName)
            # check site status
            if tmpSiteSpec.status != "online" and not site_preassigned:
                tmpLog.debug(f"  skip {tmpSiteName} due to status={tmpSiteSpec.status}")
                continue
            # check PandaSite
            if "PandaSite" in taskParamMap and taskParamMap["PandaSite"]:
                if tmpSiteSpec.pandasite != taskParamMap["PandaSite"]:
                    tmpLog.debug(f"  skip {tmpSiteName} due to wrong PandaSite={tmpSiteSpec.pandasite} <> {taskParamMap['PandaSite']}")
                    continue
            newScanSiteList.append(tmpSiteName)
        scanSiteList = newScanSiteList
        tmpLog.debug(f"{len(scanSiteList)} candidates passed site status check")
        if scanSiteList == []:
            tmpLog.error("no candidates")
            taskSpec.setErrDiag(tmpLog.uploadLog(taskSpec.jediTaskID))
            return retTmpError
        ######################################
        # selection for scratch disk
        minDiskCountS = taskSpec.getOutDiskSize() + taskSpec.getWorkDiskSize() + inputChunk.getMaxAtomSize()
        minDiskCountS = minDiskCountS // 1024 // 1024
        # size for direct IO sites
        if taskSpec.useLocalIO():
            minDiskCountR = minDiskCountS
        else:
            minDiskCountR = taskSpec.getOutDiskSize() + taskSpec.getWorkDiskSize()
            minDiskCountR = minDiskCountR // 1024 // 1024
        newScanSiteList = []
        for tmpSiteName in scanSiteList:
            tmpSiteSpec = self.siteMapper.getSite(tmpSiteName)
            # check at the site
            if tmpSiteSpec.maxwdir:
                if JediCoreUtils.use_direct_io_for_job(taskSpec, tmpSiteSpec, inputChunk):
                    minDiskCount = minDiskCountR
                else:
                    minDiskCount = minDiskCountS
                if minDiskCount > tmpSiteSpec.maxwdir:
                    tmpLog.debug(f"  skip {tmpSiteName} due to small scratch disk={tmpSiteSpec.maxwdir} < {minDiskCount}")
                    continue
            newScanSiteList.append(tmpSiteName)
        scanSiteList = newScanSiteList
        tmpLog.debug(f"{len(scanSiteList)} candidates passed scratch disk check")
        if scanSiteList == []:
            tmpLog.error("no candidates")
            taskSpec.setErrDiag(tmpLog.uploadLog(taskSpec.jediTaskID))
            return retTmpError
        ######################################
        # selection for available space in SE
        newScanSiteList = []
        for tmpSiteName in scanSiteList:
            # check at the site
            tmpSiteSpec = self.siteMapper.getSite(tmpSiteName)
            # free space must be >= 200GB
            diskThreshold = 200
            tmpSpaceSize = tmpSiteSpec.space
            if tmpSiteSpec.space and tmpSpaceSize < diskThreshold:
                tmpLog.debug(f"  skip {tmpSiteName} due to disk shortage in SE = {tmpSiteSpec.space} < {diskThreshold}GB")
                continue
            newScanSiteList.append(tmpSiteName)
        scanSiteList = newScanSiteList
        tmpLog.debug(f"{len(scanSiteList)} candidates passed SE space check")
        if not scanSiteList:
            tmpLog.error("no candidates")
            taskSpec.setErrDiag(tmpLog.uploadLog(taskSpec.jediTaskID))
            return retTmpError
        ######################################
        # selection for walltime
        minWalltime = taskSpec.walltime
        if minWalltime not in [0, None]:
            newScanSiteList = []
            for tmpSiteName in scanSiteList:
                tmpSiteSpec = self.siteMapper.getSite(tmpSiteName)
                # check at the site
                if tmpSiteSpec.maxtime != 0 and minWalltime > tmpSiteSpec.maxtime:
                    tmpLog.debug(f"  skip {tmpSiteName} due to short site walltime={tmpSiteSpec.maxtime}(site upper limit) < {minWalltime}")
                    continue
                if tmpSiteSpec.mintime != 0 and minWalltime < tmpSiteSpec.mintime:
                    tmpLog.debug(f"  skip {tmpSiteName} due to short job walltime={tmpSiteSpec.mintime}(site lower limit) > {minWalltime}")
                    continue
                newScanSiteList.append(tmpSiteName)
            scanSiteList = newScanSiteList
            tmpLog.debug(f"{len(scanSiteList)} candidates passed walltime check ={minWalltime}{taskSpec.walltimeUnit}")
            if not scanSiteList:
                tmpLog.error("no candidates")
                taskSpec.setErrDiag(tmpLog.uploadLog(taskSpec.jediTaskID))
                return retTmpError
        ######################################
        # selection for MP
        if taskSpec.coreCount is not None and taskSpec.coreCount >= 0:
            if not site_preassigned:
                newScanSiteList = []
                for tmpSiteName in scanSiteList:
                    tmpSiteSpec = self.siteMapper.getSite(tmpSiteName)
                    # check at the site
                    is_ok = False
                    if taskSpec.coreCount == 0:
                        # any
                        is_ok = True
                    elif taskSpec.coreCount == 1:
                        # score
                        if tmpSiteSpec.coreCount in [0, 1, -1, None]:
                            is_ok = True
                    else:
                        # mcore
                        if tmpSiteSpec.coreCount in [0, -1]:
                            is_ok = True
                        elif tmpSiteSpec.coreCount and tmpSiteSpec.coreCount >= taskSpec.coreCount:
                            is_ok = True
                    if is_ok:
                        newScanSiteList.append(tmpSiteName)
                    else:
                        tmpLog.info(
                            f"  skip site={tmpSiteName} due to core mismatch site:{tmpSiteSpec.coreCount} <> task:{taskSpec.coreCount} criteria=-cpucore"
                        )
                scanSiteList = newScanSiteList
                tmpLog.info(f"{len(scanSiteList)} candidates passed for core count check")
                if not scanSiteList:
                    self.dump_summary(tmpLog)
                    tmpLog.error("no candidates")
                    taskSpec.setErrDiag(tmpLog.uploadLog(taskSpec.jediTaskID))
                    return retTmpError
        ######################################
        # selection for memory
        origMinRamCount = inputChunk.getMaxRamCount()
        if not site_preassigned and origMinRamCount:
            newScanSiteList = []
            for tmpSiteName in scanSiteList:
                tmpSiteSpec = self.siteMapper.getSite(tmpSiteName)
                # job memory requirement
                if taskSpec.ramPerCore():
                    if tmpSiteSpec.coreCount and tmpSiteSpec.coreCount > 0:
                        minRamCount = origMinRamCount * tmpSiteSpec.coreCount
                    else:
                        minRamCount = origMinRamCount * (taskSpec.coreCount if taskSpec.coreCount else 1)
                    minRamCount += taskSpec.baseRamCount if taskSpec.baseRamCount else 0
                else:
                    minRamCount = origMinRamCount
                # site max memory requirement
                site_maxmemory = tmpSiteSpec.maxrss if tmpSiteSpec.maxrss else 0
                # check at the site
                if site_maxmemory and minRamCount and minRamCount > site_maxmemory:
                    tmpMsg = f"  skip site={tmpSiteName} due to site RAM shortage. {site_maxmemory} (site upper limit) less than {minRamCount} "
                    tmpLog.debug(tmpMsg)
                    continue
                # site min memory requirement
                site_minmemory = tmpSiteSpec.minrss if tmpSiteSpec.minrss else 0
                if site_minmemory and minRamCount and minRamCount < site_minmemory:
                    tmpMsg = f"  skip site={tmpSiteName} due to job RAM shortage. {site_minmemory} (site lower limit) greater than {minRamCount} "
                    tmpLog.info(tmpMsg)
                    continue
                newScanSiteList.append(tmpSiteName)
            scanSiteList = newScanSiteList
            tmpLog.debug(f"{len(scanSiteList)} candidates passed memory check")
            if not scanSiteList:
                tmpLog.error("no candidates")
                taskSpec.setErrDiag(tmpLog.uploadLog(taskSpec.jediTaskID))
                return retTmpError
        ######################################
        # selection for nPilot
        nWNmap = self.taskBufferIF.getCurrentSiteData()
        newScanSiteList = []
        for tmpSiteName in scanSiteList:
            # check at the site
            nPilot = 0
            if tmpSiteName in nWNmap:
                nPilot = nWNmap[tmpSiteName]["getJob"] + nWNmap[tmpSiteName]["updateJob"]
            if nPilot == 0 and taskSpec.prodSourceLabel not in ["test"]:
                tmpLog.debug(f"  skip {tmpSiteName} due to no pilot")
                # continue
            newScanSiteList.append(tmpSiteName)
        scanSiteList = newScanSiteList
        tmpLog.debug(f"{len(scanSiteList)} candidates passed pilot activity check")
        if scanSiteList == []:
            tmpLog.error("no candidates")
            taskSpec.setErrDiag(tmpLog.uploadLog(taskSpec.jediTaskID))
            return retTmpError
        ######################################
        # sites already used by task
        tmpSt, sitesUsedByTask = self.taskBufferIF.getSitesUsedByTask_JEDI(taskSpec.jediTaskID)
        if not tmpSt:
            tmpLog.error("failed to get sites which already used by task")
            taskSpec.setErrDiag(tmpLog.uploadLog(taskSpec.jediTaskID))
            return retTmpError
        ######################################
        # get list of available files
        availableFileMap = {}
        for datasetSpec in inputChunk.getDatasets():
            try:
                # get list of site to be scanned
                tmpLog.debug(f"getting the list of available files for {datasetSpec.datasetName}")
                fileScanSiteList = []
                for tmpPseudoSiteName in scanSiteList:
                    tmpSiteSpec = self.siteMapper.getSite(tmpPseudoSiteName)
                    tmpSiteName = tmpSiteSpec.get_unified_name()
                    if tmpSiteName in fileScanSiteList:
                        continue
                    fileScanSiteList.append(tmpSiteName)
                # mapping between sites and input storage endpoints
                siteStorageEP = AtlasBrokerUtils.getSiteInputStorageEndpointMap(fileScanSiteList, self.siteMapper, taskSpec.prodSourceLabel, None)
                # disable file lookup for merge jobs
                if inputChunk.isMerging:
                    checkCompleteness = False
                else:
                    checkCompleteness = True
                if not datasetSpec.isMaster():
                    useCompleteOnly = True
                else:
                    useCompleteOnly = False
                # get available files per site/endpoint
                tmpAvFileMap = self.ddmIF.getAvailableFiles(
                    datasetSpec,
                    siteStorageEP,
                    self.siteMapper,
                    check_completeness=checkCompleteness,
                    file_scan_in_container=False,
                    complete_only=useCompleteOnly,
                    use_deep=True,
                )
                if tmpAvFileMap is None:
                    raise Interaction.JEDITemporaryError("ddmIF.getAvailableFiles failed")
                availableFileMap[datasetSpec.datasetName] = tmpAvFileMap
            except Exception as e:
                tmpLog.error(f"failed to get available files with {e}")
                taskSpec.setErrDiag(tmpLog.uploadLog(taskSpec.jediTaskID))
                return retTmpError
        ######################################
        # calculate weight
        tmpSt, jobStatPrioMap = self.taskBufferIF.getJobStatisticsByGlobalShare(taskSpec.vo)
        if not tmpSt:
            tmpLog.error("failed to get job statistics with priority")
            taskSpec.setErrDiag(tmpLog.uploadLog(taskSpec.jediTaskID))
            return retTmpError
        ######################################
        # final procedure
        tmpLog.debug(f"final {len(scanSiteList)} candidates")
        weightMap = {}
        candidateSpecList = []
        preSiteCandidateSpec = None
        for tmpSiteName in scanSiteList:
            # get number of jobs in each job status. Using workQueueID=None to include non-JEDI jobs
            nRunning = AtlasBrokerUtils.getNumJobs(jobStatPrioMap, tmpSiteName, "running", None, None)
            nAssigned = AtlasBrokerUtils.getNumJobs(jobStatPrioMap, tmpSiteName, "defined", None, None)
            nActivated = AtlasBrokerUtils.getNumJobs(jobStatPrioMap, tmpSiteName, "activated", None, None)
            weight = float(nRunning + 1) / float(nActivated + nAssigned + 1) / float(nAssigned + 1)
            # make candidate
            siteCandidateSpec = SiteCandidate(tmpSiteName)
            # set weight
            siteCandidateSpec.weight = weight
            # files
            for tmpDatasetName, availableFiles in availableFileMap.items():
                if tmpSiteName in availableFiles:
                    siteCandidateSpec.add_local_disk_files(availableFiles[tmpSiteName]["localdisk"])
            # append
            if tmpSiteName in sitesUsedByTask:
                candidateSpecList.append(siteCandidateSpec)
            else:
                if weight not in weightMap:
                    weightMap[weight] = []
                weightMap[weight].append(siteCandidateSpec)
        # limit the number of sites
        maxNumSites = 5
        weightList = sorted(weightMap.keys())
        weightList.reverse()
        for weightVal in weightList:
            if len(candidateSpecList) >= maxNumSites:
                break
            sitesWithWeight = weightMap[weightVal]
            random.shuffle(sitesWithWeight)
            candidateSpecList += sitesWithWeight[: (maxNumSites - len(candidateSpecList))]
        # collect site names
        scanSiteList = []
        for siteCandidateSpec in candidateSpecList:
            scanSiteList.append(siteCandidateSpec.siteName)
        # append candidates
        newScanSiteList = []
        for siteCandidateSpec in candidateSpecList:
            # append
            inputChunk.addSiteCandidate(siteCandidateSpec)
            newScanSiteList.append(siteCandidateSpec.siteName)
            tmpLog.debug(f"  use {siteCandidateSpec.siteName} with weight={siteCandidateSpec.weight} nFiles={len(siteCandidateSpec.localDiskFiles)}")
        scanSiteList = newScanSiteList
        if scanSiteList == []:
            tmpLog.error("no candidates")
            taskSpec.setErrDiag(tmpLog.uploadLog(taskSpec.jediTaskID))
            return retTmpError
        # return
        tmpLog.debug("done")
        return self.SC_SUCCEEDED, inputChunk



import datetime
import multiprocessing
import sys

from . import JediCoreUtils


# wrapper for multiprocessing.Process
class ProcessWrapper(multiprocessing.Process):
    # constructor
    def __init__(self, target, args):
        multiprocessing.Process.__init__(self, target=self.wrappedMain)
        self.target = target
        self.args = args

    # main
    def wrappedMain(self):
        while True:
            proc = multiprocessing.Process(target=self.target, args=self.args)
            proc.start()
            pid = proc.pid
            while True:
                try:
                    proc.join(20)
                    if not JediCoreUtils.checkProcess(pid):
                        timeNow = datetime.datetime.now(datetime.timezone.utc).replace(tzinfo=None)
                        print(f"{str(timeNow)} {self.__class__.__name__}: INFO    pid={pid} not exist")
                        break
                except Exception:
                    timeNow = datetime.datetime.now(datetime.timezone.utc).replace(tzinfo=None)
                    errType, errValue = sys.exc_info()[:2]
                    print(f"{str(timeNow)} {self.__class__.__name__}: INFO    failed to check pid={pid} with {errType} {errValue}")



import datetime
import grp
import multiprocessing
import optparse
import os
import pwd
import signal
import sys
import time

import daemon
from pandajedi.jediconfig import jedi_config
from pandajedi.jedicore.JediTaskBufferInterface import JediTaskBufferInterface
from pandajedi.jedicore.ProcessUtils import ProcessWrapper
from pandajedi.jedicore.ThreadUtils import ZombieCleaner
from pandajedi.jediddm.DDMInterface import DDMInterface


# the master class of JEDI which runs the main process
class JediMaster:
    # constrictor
    def __init__(self):
        self.stopEventList = []

    # spawn a knight to have own file descriptors
    def launcher(self, moduleName, *args, **kwargs):
        # import module
        mod = __import__(moduleName)
        for subModuleName in moduleName.split(".")[1:]:
            mod = getattr(mod, subModuleName)
        # launch
        timeNow = datetime.datetime.now(datetime.timezone.utc).replace(tzinfo=None)
        print(f"{str(timeNow)} {moduleName}: INFO    start launcher with pid={os.getpid()}")
        mod.launcher(*args, **kwargs)

    # convert config parameters
    def convParams(self, itemStr):
        items = itemStr.split(":")
        newItems = []
        for item in items:
            if item == "":
                newItems.append(None)
            elif "," in item:
                newItems.append(item.split(","))
            else:
                try:
                    newItems.append(int(item))
                except Exception:
                    newItems.append(item)
        return newItems

    # main loop
    def start(self):
        # start zombi cleaner
        ZombieCleaner().start()
        # setup DDM I/F
        ddmIF = DDMInterface()
        ddmIF.setupInterface()
        # setup TaskBuffer I/F
        taskBufferIF = JediTaskBufferInterface()
        taskBufferIF.setupInterface()
        # the list of JEDI knights
        knightList = []
        # setup TaskRefiner
        for itemStr in jedi_config.taskrefine.procConfig.split(";"):
            items = self.convParams(itemStr)
            vo = items[0]
            plabel = items[1]
            nProc = items[2]
            for iproc in range(nProc):
                parent_conn, child_conn = multiprocessing.Pipe()
                proc = multiprocessing.Process(target=self.launcher, args=("pandajedi.jediorder.TaskRefiner", child_conn, taskBufferIF, ddmIF, vo, plabel))
                proc.start()
                knightList.append(proc)
        # setup TaskBrokerage
        for itemStr in jedi_config.taskbroker.procConfig.split(";"):
            items = self.convParams(itemStr)
            vo = items[0]
            plabel = items[1]
            nProc = items[2]
            for iproc in range(nProc):
                parent_conn, child_conn = multiprocessing.Pipe()
                proc = multiprocessing.Process(target=self.launcher, args=("pandajedi.jediorder.TaskBroker", child_conn, taskBufferIF, ddmIF, vo, plabel))
                proc.start()
                knightList.append(proc)
        # setup ContentsFeeder
        for itemStr in jedi_config.confeeder.procConfig.split(";"):
            items = self.convParams(itemStr)
            vo = items[0]
            plabel = items[1]
            nProc = items[2]
            for iproc in range(nProc):
                parent_conn, child_conn = multiprocessing.Pipe()
                proc = multiprocessing.Process(target=self.launcher, args=("pandajedi.jediorder.ContentsFeeder", child_conn, taskBufferIF, ddmIF, vo, plabel))
                proc.start()
                knightList.append(proc)
        # setup JobGenerator
        for itemStr in jedi_config.jobgen.procConfig.split(";"):
            items = self.convParams(itemStr)
            vo = items[0]
            plabel = items[1]
            nProc = items[2]
            cloud = items[3]
            try:
                loop_cycle = items[4]
            except IndexError:
                loop_cycle = None

            if not isinstance(cloud, list):
                cloud = [cloud]
            for iproc in range(nProc):
                parent_conn, child_conn = multiprocessing.Pipe()
                proc = ProcessWrapper(
                    target=self.launcher, args=("pandajedi.jediorder.JobGenerator", child_conn, taskBufferIF, ddmIF, vo, plabel, cloud, True, True, loop_cycle)
                )
                proc.start()
                knightList.append(proc)
        # setup PostProcessor
        for itemStr in jedi_config.postprocessor.procConfig.split(";"):
            items = self.convParams(itemStr)
            vo = items[0]
            plabel = items[1]
            nProc = items[2]
            for iproc in range(nProc):
                parent_conn, child_conn = multiprocessing.Pipe()
                proc = multiprocessing.Process(target=self.launcher, args=("pandajedi.jediorder.PostProcessor", child_conn, taskBufferIF, ddmIF, vo, plabel))
                proc.start()
                knightList.append(proc)
        # setup TaskCommando
        for itemStr in jedi_config.tcommando.procConfig.split(";"):
            items = self.convParams(itemStr)
            vo = items[0]
            plabel = items[1]
            nProc = items[2]
            for iproc in range(nProc):
                parent_conn, child_conn = multiprocessing.Pipe()
                proc = multiprocessing.Process(target=self.launcher, args=("pandajedi.jediorder.TaskCommando", child_conn, taskBufferIF, ddmIF, vo, plabel))
                proc.start()
                knightList.append(proc)
        # setup WatchDog
        for itemStr in jedi_config.watchdog.procConfig.split(";"):
            items = self.convParams(itemStr)
            vo = items[0]
            plabel = items[1]
            nProc = items[2]
            subStr = items[3] if len(items) > 3 else None
            period = items[4] if len(items) > 4 else None
            for iproc in range(nProc):
                parent_conn, child_conn = multiprocessing.Pipe()
                proc = multiprocessing.Process(
                    target=self.launcher, args=("pandajedi.jediorder.WatchDog", child_conn, taskBufferIF, ddmIF, vo, plabel, subStr, period)
                )
                proc.start()
                knightList.append(proc)
        # setup JediMsgProcessor agent (only one system process)
        if hasattr(jedi_config, "msgprocessor") and hasattr(jedi_config.msgprocessor, "configFile") and jedi_config.msgprocessor.configFile:
            stop_event = multiprocessing.Event()
            self.stopEventList.append(stop_event)
            parent_conn, child_conn = multiprocessing.Pipe()
            proc = multiprocessing.Process(target=self.launcher, args=("pandajedi.jediorder.JediMsgProcessor", stop_event))
            proc.start()
            knightList.append(proc)
        # setup JediDaemon agent (only one system process)
        if hasattr(jedi_config, "daemon") and hasattr(jedi_config.daemon, "enable") and jedi_config.daemon.enable:
            parent_conn, child_conn = multiprocessing.Pipe()
            proc = multiprocessing.Process(target=self.launcher, args=("pandajedi.jediorder.JediDaemon", taskBufferIF, ddmIF))
            proc.start()
            knightList.append(proc)
        # check initial failures
        time.sleep(5)
        for knight in knightList:
            if not knight.is_alive():
                timeNow = datetime.datetime.now(datetime.timezone.utc).replace(tzinfo=None)
                print(f"{str(timeNow)} {self.__class__.__name__}: ERROR    pid={knight.pid} died in initialization")
                os.killpg(os.getpgrp(), signal.SIGKILL)
        # join
        for knight in knightList:
            knight.join()

    # graceful stop
    def stop(self):
        for stop_event in self.stopEventList:
            stop_event.set()


# kill whole process
def kill_whole(sig, frame):
    # kill
    os.killpg(os.getpgrp(), signal.SIGKILL)


# main
if __name__ == "__main__":
    # parse option
    parser = optparse.OptionParser()
    parser.add_option("--pid", action="store", dest="pid", default=None, help="pid filename")
    options, args = parser.parse_args()
    uid = None
    gid = None
    if "PANDA_NO_ROOT" not in os.environ:
        if jedi_config.master.uname:
            uid = pwd.getpwnam(jedi_config.master.uname).pw_uid
        if jedi_config.master.gname:
            gid = grp.getgrnam(jedi_config.master.gname).gr_gid
    timeNow = datetime.datetime.now(datetime.timezone.utc).replace(tzinfo=None)
    print(f"{str(timeNow)} JediMaster: INFO    start")
    # make daemon context
    dc = daemon.DaemonContext(stdout=sys.stdout, stderr=sys.stderr, uid=uid, gid=gid)
    with dc:
        # record PID
        go_ahead = True
        try:
            if options.pid:  # pid files are no longer necessary in systemd
                pidFile = open(options.pid, "x")
        except FileExistsError:
            print(f"{str(timeNow)} JediMaster: ERROR    terminated since pid file {options.pid} already exists")
            go_ahead = False
        if go_ahead:
            if options.pid:  # pid files are no longer necessary in systemd
                pidFile.write(f"{os.getpid()}")
                pidFile.close()

            # master
            master = JediMaster()

            # set handler
            def catch_sig(sig, frame):
                master.stop()
                time.sleep(3)
                kill_whole(sig, frame)

            signal.signal(signal.SIGINT, catch_sig)
            signal.signal(signal.SIGHUP, catch_sig)
            signal.signal(signal.SIGTERM, catch_sig)
            # start master
            master.start()



import sys

try:
    testTaskType = sys.argv[1]
except Exception:
    testTaskType = "test"

try:
    vo = sys.argv[2]
except Exception:
    vo = "atlas"

import multiprocessing

from pandajedi.jedicore.JediTaskBufferInterface import JediTaskBufferInterface
from pandajedi.jediddm.DDMInterface import DDMInterface
from pandajedi.jediorder import TaskRefiner

tbIF = JediTaskBufferInterface()
tbIF.setupInterface()


ddmIF = DDMInterface()
ddmIF.setupInterface()


parent_conn, child_conn = multiprocessing.Pipe()

taskRefiner = multiprocessing.Process(target=TaskRefiner.launcher, args=(child_conn, tbIF, ddmIF, vo, testTaskType))
taskRefiner.start()



import json
import re
import sys
import uuid

from pandacommon.pandalogger.PandaLogger import PandaLogger
from pandajedi.jedicore import Interaction
from pandajedi.jedicore.MsgWrapper import MsgWrapper

from .TaskGeneratorBase import TaskGeneratorBase

logger = PandaLogger().getLogger(__name__.split(".")[-1])


# task generator for ATLAS
class AtlasTaskGenerator(TaskGeneratorBase):
    # constructor
    def __init__(self, taskBufferIF, ddmIF):
        TaskGeneratorBase.__init__(self, taskBufferIF, ddmIF)

    # main to generate task
    def doGenerate(self, taskSpec, taskParamMap, **varMap):
        # make logger
        tmpLog = MsgWrapper(logger, f"<jediTaskID={taskSpec.jediTaskID}>")
        tmpLog.info(f"start taskType={taskSpec.taskType}")
        tmpLog.info(str(varMap))
        # returns
        retFatal = self.SC_FATAL
        retTmpError = self.SC_FAILED
        retOK = self.SC_SUCCEEDED
        try:
            # check prodSourceLabel
            if taskSpec.prodSourceLabel in ["managed", "test"]:
                # check taskType
                if taskSpec.taskType == "recov":
                    # generate parent tasks for lost file recovery if it is not yet generated
                    if "parentGenerated" in taskParamMap:
                        tmpLog.info("skip since already generated parent tasks")
                    else:
                        tmpLog.info("generating parent tasks for lost file recovery")
                        # missing files are undefined
                        if "missingFilesMap" not in varMap:
                            tmpLog.error("missing files are undefined")
                            return retFatal
                        missingFilesMap = varMap["missingFilesMap"]
                        # check datasets
                        for datasetName, datasetValMap in missingFilesMap.items():
                            # dataset needs specify container
                            datasetSpec = datasetValMap["datasetSpec"]
                            if datasetSpec.containerName in ["", None]:
                                errStr = f"cannot make parent tasks due to undefined container for datasetID={datasetSpec.datasetID}:{datasetName}"
                                taskSpec.setErrDiag(errStr)
                                tmpLog.error(errStr)
                                return retFatal
                        # make parameters for new task
                        newJsonStrList = []
                        for datasetName, datasetValMap in missingFilesMap.items():
                            datasetSpec = datasetValMap["datasetSpec"]
                            newTaskParamMap = {}
                            newTaskParamMap["oldDatasetName"] = datasetName
                            newTaskParamMap["lostFiles"] = datasetValMap["missingFiles"]
                            newTaskParamMap["vo"] = taskSpec.vo
                            newTaskParamMap["cloud"] = taskSpec.cloud
                            newTaskParamMap["taskPriority"] = taskSpec.taskPriority
                            newTaskParamMap["taskType"] = taskSpec.taskType
                            newTaskParamMap["prodSourceLabel"] = taskSpec.prodSourceLabel
                            logDatasetName = f"panda.jedi{taskSpec.taskType}.log.{uuid.uuid4()}"
                            newTaskParamMap["log"] = {
                                "dataset": logDatasetName,
                                "type": "template",
                                "param_type": "log",
                                "token": "ATLASDATADISK",
                                "value": f"{logDatasetName}.${{SN}}.log.tgz",
                            }
                            # make new datasetname
                            outDatasetName = datasetName
                            # remove /
                            outDatasetName = re.sub("/$", "", outDatasetName)
                            # remove extension
                            outDatasetName = re.sub(f"\\.{taskSpec.taskType}\\d+$", "", outDatasetName)
                            # add extension
                            outDatasetName = outDatasetName + f".{taskSpec.taskType}{taskSpec.jediTaskID}"
                            newTaskParamMap["output"] = {"dataset": outDatasetName}
                            if datasetSpec.containerName not in ["", None]:
                                newTaskParamMap["output"]["container"] = datasetSpec.containerName
                            # make json
                            jsonStr = json.dumps(newTaskParamMap)
                            newJsonStrList.append(jsonStr)
                        # change original task parameters to not repeat the same procedure and to use newly produced files
                        taskParamMap["parentGenerated"] = True
                        taskParamMap["useInFilesInContainer"] = True
                        taskParamMap["useInFilesWithNewAttemptNr"] = True
                        jsonStr = json.dumps(taskParamMap)
                        # insert and update task parameters
                        sTmp, newJediTaskIDs = self.taskBufferIF.insertUpdateTaskParams_JEDI(
                            taskSpec.jediTaskID, taskSpec.vo, taskSpec.prodSourceLabel, jsonStr, newJsonStrList
                        )
                        if sTmp:
                            tmpLog.info(f"inserted/updated tasks in DB : new jediTaskIDs={str(newJediTaskIDs)}")
                        else:
                            tmpLog.error("failed to insert/update tasks in DB")
                            return retFatal
            # return
            tmpLog.info("done")
            return retOK
        except Exception:
            errtype, errvalue = sys.exc_info()[:2]
            tmpLog.error(f"doGenerate failed with {errtype.__name__}:{errvalue}")
            return retFatal



# logger
from pandacommon.pandalogger.PandaLogger import PandaLogger
from pandajedi.jediconfig import jedi_config
from pandajedi.jedicore.FactoryBase import FactoryBase

logger = PandaLogger().getLogger(__name__.split(".")[-1])


# factory class for job brokerage
class JobBroker(FactoryBase):
    # constructor
    def __init__(self, vo, sourceLabel):
        FactoryBase.__init__(self, vo, sourceLabel, logger, jedi_config.jobbroker.modConfig)

    # main
    def doBrokerage(self, taskSpec, cloudName, inputChunk, taskParamMap):
        return self.getImpl(taskSpec.vo, taskSpec.prodSourceLabel).doBrokerage(taskSpec, cloudName, inputChunk, taskParamMap)

    # set live counter
    def setLiveCounter(self, vo, sourceLabel, liveCounter):
        self.getImpl(vo, sourceLabel).setLiveCounter(liveCounter)

    # set lock ID
    def setLockID(self, vo, sourceLabel, pid, tid):
        self.getImpl(vo, sourceLabel).setLockID(pid, tid)

    # get base lock ID
    def getBaseLockID(self, vo, sourceLabel):
        return self.getImpl(vo, sourceLabel).getBaseLockID()

    # set test mode
    def setTestMode(self, vo, sourceLabel):
        self.getImpl(vo, sourceLabel).setTestMode()



import sys
import traceback

from pandacommon.pandalogger.PandaLogger import PandaLogger
from pandajedi.jedicore.MsgWrapper import MsgWrapper
from pandaserver.dataservice import DataServiceUtils
from pandaserver.taskbuffer import EventServiceUtils, JobUtils

from .TaskSetupperBase import TaskSetupperBase

logger = PandaLogger().getLogger(__name__.split(".")[-1])


# task setup for ATLAS
class AtlasTaskSetupper(TaskSetupperBase):
    # constructor
    def __init__(self, taskBufferIF, ddmIF):
        TaskSetupperBase.__init__(self, taskBufferIF, ddmIF)

    # main to setup task
    def doSetup(self, taskSpec, datasetToRegister, pandaJobs):
        # make logger
        tmpLog = MsgWrapper(logger, f"< jediTaskID={taskSpec.jediTaskID} >")
        tmpLog.info(f"start label={taskSpec.prodSourceLabel} taskType={taskSpec.taskType}")
        # returns
        retFatal = self.SC_FATAL
        retTmpError = self.SC_FAILED
        retOK = self.SC_SUCCEEDED
        try:
            # get DDM I/F
            ddmIF = self.ddmIF.getInterface(taskSpec.vo)
            # register datasets
            if datasetToRegister != [] or taskSpec.prodSourceLabel in ["user"]:
                # prod vs anal
                userSetup = False
                if taskSpec.prodSourceLabel in ["user"]:
                    userSetup = True
                    # collect datasetID to register datasets/containers just in case
                    for tmpPandaJob in pandaJobs:
                        if not tmpPandaJob.produceUnMerge():
                            for tmpFileSpec in tmpPandaJob.Files:
                                if tmpFileSpec.type in ["output", "log"]:
                                    if tmpFileSpec.datasetID not in datasetToRegister:
                                        datasetToRegister.append(tmpFileSpec.datasetID)
                tmpLog.info(f"datasetToRegister={str(datasetToRegister)}")
                # get site mapper
                siteMapper = self.taskBufferIF.getSiteMapper()

                # loop over all datasets
                avDatasetList = []
                cnDatasetMap = {}
                for datasetID in datasetToRegister:
                    # get output and log datasets
                    tmpLog.info(f"getting datasetSpec with datasetID={datasetID}")
                    tmpStat, datasetSpec = self.taskBufferIF.getDatasetWithID_JEDI(taskSpec.jediTaskID, datasetID)
                    if not tmpStat:
                        tmpLog.error("failed to get output and log datasets")
                        return retFatal
                    if datasetSpec.isPseudo():
                        tmpLog.info("skip pseudo dataset")
                        continue
                    # DDM backend
                    ddmBackEnd = taskSpec.getDdmBackEnd()
                    tmpLog.info(f"checking {datasetSpec.datasetName}")
                    # secondary nucleus
                    nucleusSpec = siteMapper.getNucleus(taskSpec.nucleus)
                    secNucleusSpecBase = None
                    if nucleusSpec and not taskSpec.is_intermediate_task():
                        secondaryNucleus = nucleusSpec.get_secondary_nucleus()
                        if secondaryNucleus:
                            secNucleusSpecBase = siteMapper.getNucleus(secondaryNucleus)
                    # check if dataset and container are available in DDM
                    for targetName in [datasetSpec.datasetName, datasetSpec.containerName]:
                        if targetName is None:
                            continue
                        if targetName not in avDatasetList:
                            # set lifetime
                            if targetName.startswith("panda"):
                                if datasetSpec.type == "trn_log" and taskSpec.prodSourceLabel == "managed":
                                    lifetime = 365
                                else:
                                    lifetime = 14
                            else:
                                lifetime = None
                            # check dataset/container in DDM
                            tmpList = ddmIF.listDatasets(targetName)
                            if tmpList == []:
                                # get location
                                location = None
                                locForRule = None
                                if targetName == datasetSpec.datasetName:
                                    # dataset
                                    if datasetSpec.site in ["", None]:
                                        if DataServiceUtils.getDistributedDestination(datasetSpec.storageToken) is not None:
                                            locForRule = datasetSpec.destination
                                        elif DataServiceUtils.getDestinationSE(datasetSpec.storageToken) is not None:
                                            location = DataServiceUtils.getDestinationSE(datasetSpec.storageToken)
                                        elif taskSpec.cloud is not None:
                                            # use T1 SE
                                            tmpT1Name = siteMapper.getCloud(taskSpec.cloud)["source"]
                                            location = siteMapper.getDdmEndpoint(
                                                tmpT1Name,
                                                datasetSpec.storageToken,
                                                taskSpec.prodSourceLabel,
                                                JobUtils.translate_tasktype_to_jobtype(taskSpec.taskType),
                                            )
                                    else:
                                        tmpLog.info(f"site={datasetSpec.site} token={datasetSpec.storageToken}")
                                        location = siteMapper.getDdmEndpoint(
                                            datasetSpec.site,
                                            datasetSpec.storageToken,
                                            taskSpec.prodSourceLabel,
                                            JobUtils.translate_tasktype_to_jobtype(taskSpec.taskType),
                                        )
                                if locForRule is None:
                                    locForRule = location
                                # set metadata
                                if taskSpec.prodSourceLabel in ["managed", "test"] and targetName == datasetSpec.datasetName:
                                    metaData = {}
                                    metaData["task_id"] = taskSpec.jediTaskID
                                    if taskSpec.campaign not in [None, ""]:
                                        metaData["campaign"] = taskSpec.campaign
                                    if datasetSpec.getTransient() is not None:
                                        metaData["transient"] = datasetSpec.getTransient()
                                else:
                                    metaData = None
                                # use secondary nucleus only for production output
                                if (
                                    secNucleusSpecBase
                                    and taskSpec.prodSourceLabel in ["managed", "test"]
                                    and targetName == datasetSpec.datasetName
                                    and datasetSpec.type == "output"
                                ):
                                    secNucleusSpec = secNucleusSpecBase
                                else:
                                    secNucleusSpec = None
                                # register dataset/container
                                tmpLog.info(f"registering {targetName} with location={location} backend={ddmBackEnd} lifetime={lifetime} meta={str(metaData)}")
                                tmpStat = ddmIF.registerNewDataset(targetName, backEnd=ddmBackEnd, location=location, lifetime=lifetime, metaData=metaData)
                                if not tmpStat:
                                    tmpLog.error(f"failed to register {targetName}")
                                    return retFatal
                                # procedures for user
                                if userSetup or DataServiceUtils.getDistributedDestination(datasetSpec.storageToken) is not None or secNucleusSpec:
                                    # register location
                                    tmpToRegister = False
                                    if userSetup and targetName == datasetSpec.datasetName and datasetSpec.site not in ["", None]:
                                        if taskSpec.workingGroup:
                                            userName = taskSpec.workingGroup
                                        else:
                                            userName = taskSpec.userName
                                        grouping = None
                                        tmpToRegister = True
                                    elif DataServiceUtils.getDistributedDestination(datasetSpec.storageToken) is not None:
                                        userName = None
                                        grouping = "NONE"
                                        tmpToRegister = True
                                    elif secNucleusSpec:
                                        userName = None
                                        grouping = None
                                        tmpToRegister = True
                                        locForRule = siteMapper.getDdmEndpoint(
                                            secNucleusSpec.getOnePandaSite(),
                                            datasetSpec.storageToken,
                                            taskSpec.prodSourceLabel,
                                            JobUtils.translate_tasktype_to_jobtype(taskSpec.taskType),
                                        )

                                    if tmpToRegister:
                                        activity = DataServiceUtils.getActivityForOut(taskSpec.prodSourceLabel)
                                        tmpLog.info(
                                            "registering location={} lifetime={} days activity={} grouping={} "
                                            "owner={}".format(locForRule, lifetime, activity, grouping, userName)
                                        )
                                        tmpStat = ddmIF.registerDatasetLocation(
                                            targetName, locForRule, owner=userName, lifetime=lifetime, backEnd=ddmBackEnd, activity=activity, grouping=grouping
                                        )
                                        if not tmpStat:
                                            tmpLog.error(f"failed to register location {locForRule} for {targetName}")
                                            return retFatal
                                        # double copy
                                        if userSetup and datasetSpec.type == "output":
                                            if datasetSpec.destination != datasetSpec.site:
                                                tmpLog.info(f"skip making double copy as destination={datasetSpec.destination} is not site={datasetSpec.site}")
                                            else:
                                                second_copy = True
                                                try:
                                                    if taskSpec.site:
                                                        panda_site = siteMapper.getSite(taskSpec.site)
                                                        if panda_site.catchall and "skip_2nd_copy" in panda_site.catchall:
                                                            tmpLog.info(f"skip making double copy as specified in {panda_site} catchall")
                                                            second_copy = False
                                                except Exception:
                                                    second_copy = True

                                                if second_copy:
                                                    locForDouble = f"(type=SCRATCHDISK)\\notforextracopy=True\\{locForRule}"
                                                    tmpMsg = "registering double copy "
                                                    tmpMsg += f'location="{locForDouble}" lifetime={lifetime}days activity={activity} for dataset={targetName}'
                                                    tmpLog.info(tmpMsg)
                                                    tmpStat = ddmIF.registerDatasetLocation(
                                                        targetName,
                                                        locForDouble,
                                                        copies=1,
                                                        owner=userName,
                                                        lifetime=lifetime,
                                                        activity=activity,
                                                        grouping="NONE",
                                                        weight="freespace",
                                                        ignore_availability=False,
                                                    )
                                                    if not tmpStat:
                                                        tmpLog.error(f"failed to register double copylocation {locForDouble} for {targetName}")
                                                        return retFatal
                                avDatasetList.append(targetName)
                            else:
                                tmpLog.info(f"{targetName} already registered")
                    # check if dataset is in the container
                    if datasetSpec.containerName is not None and datasetSpec.containerName != datasetSpec.datasetName:
                        # get list of constituent datasets in the container
                        if datasetSpec.containerName not in cnDatasetMap:
                            cnDatasetMap[datasetSpec.containerName] = ddmIF.listDatasetsInContainer(datasetSpec.containerName)
                        # add dataset
                        if datasetSpec.datasetName not in cnDatasetMap[datasetSpec.containerName]:
                            tmpLog.info(f"adding {datasetSpec.datasetName} to {datasetSpec.containerName}")
                            tmpStat = ddmIF.addDatasetsToContainer(datasetSpec.containerName, [datasetSpec.datasetName], backEnd=ddmBackEnd)
                            if not tmpStat:
                                tmpLog.error(f"failed to add {datasetSpec.datasetName} to {datasetSpec.containerName}")
                                return retFatal
                            cnDatasetMap[datasetSpec.containerName].append(datasetSpec.datasetName)
                        else:
                            tmpLog.info(f"{datasetSpec.datasetName} already in {datasetSpec.containerName}")
                    # update dataset
                    datasetSpec.status = "registered"
                    self.taskBufferIF.updateDataset_JEDI(datasetSpec, {"jediTaskID": taskSpec.jediTaskID, "datasetID": datasetID})
            # register ES datasets
            if taskSpec.registerEsFiles():
                targetName = EventServiceUtils.getEsDatasetName(taskSpec.jediTaskID)
                location = None
                metaData = {}
                metaData["task_id"] = taskSpec.jediTaskID
                metaData["hidden"] = True
                tmpLog.info(f"registering ES dataset {targetName} with location={location} meta={str(metaData)}")
                tmpStat = ddmIF.registerNewDataset(targetName, location=location, metaData=metaData, resurrect=True)
                if not tmpStat:
                    tmpLog.error(f"failed to register ES dataset {targetName}")
                    return retFatal
                # register rule
                location = "type=DATADISK"
                activity = DataServiceUtils.getActivityForOut(taskSpec.prodSourceLabel)
                grouping = "NONE"
                tmpLog.info(f"registering location={location} activity={activity} grouping={grouping}")
                tmpStat = ddmIF.registerDatasetLocation(targetName, location, activity=activity, grouping=grouping)
                if not tmpStat:
                    tmpLog.error(f"failed to register location {location} with {activity} for {targetName}")
                    return retFatal
            # open datasets
            if taskSpec.prodSourceLabel in ["managed", "test"]:
                # get the list of output/log datasets
                outDatasetList = []
                for tmpPandaJob in pandaJobs:
                    for tmpFileSpec in tmpPandaJob.Files:
                        if tmpFileSpec.type in ["output", "log"]:
                            if tmpFileSpec.destinationDBlock not in outDatasetList:
                                outDatasetList.append(tmpFileSpec.destinationDBlock)
                # open datasets
                for outDataset in outDatasetList:
                    tmpLog.info(f"open {outDataset}")
                    ddmIF.openDataset(outDataset)
                    # unset lifetime
                    ddmIF.setDatasetMetadata(outDataset, "lifetime", None)
            # return
            tmpLog.info("done")
            return retOK
        except Exception as e:
            tmpLog.error(f"doSetup failed with {str(e)}")
            tb = traceback.format_exc()
            taskSpec.setErrDiag(tmpLog.uploadLog(taskSpec.jediTaskID))
            tmpLog.error(tb)
            return retFatal



import sys

try:
    metaID = sys.argv[1]
except Exception:
    metaID = None
import json
import uuid

taskParamMap = {}
taskParamMap["taskName"] = str(uuid.uuid4())
taskParamMap["userName"] = "pandasrv1"
taskParamMap["vo"] = "atlas"
taskParamMap["taskPriority"] = 100
taskParamMap["architecture"] = "i686-slc5-gcc43-opt"
taskParamMap["transUses"] = "Atlas-17.2.7"
taskParamMap["transHome"] = "AtlasProduction-17.2.8.10"
taskParamMap["transPath"] = "Reco_trf.py"
taskParamMap["processingType"] = "reco"
taskParamMap["prodSourceLabel"] = "test"
taskParamMap["taskType"] = "prod"
taskParamMap["workingGroup"] = "AP_Higgs"
taskParamMap["coreCount"] = 1
taskParamMap["cloud"] = "US"
logDatasetName = f"panda.jeditest.log.{uuid.uuid4()}"
taskParamMap["log"] = {
    "dataset": logDatasetName,
    "type": "template",
    "param_type": "log",
    "token": "ATLASDATADISK",
    "value": f"{logDatasetName}.${{SN}}.log.tgz",
}
outDatasetName = f"panda.jeditest.NTUP_EMBLLDN.{uuid.uuid4()}"
taskParamMap["jobParameters"] = [
    {
        "type": "template",
        "param_type": "input",
        "value": "inputAODFile=${IN}",
        "dataset": "data12_8TeV.00214651.physics_Egamma.merge.AOD.f489_m1261",
    },
    {
        "type": "constant",
        "value": 'maxEvents=1000 RunNumber=213816 autoConfiguration=everything preExec="from BTagging.BTaggingFlags import BTaggingFlags;BTaggingFlags.CalibrationTag="BTagCalibALL-07-02""',
    },
    {
        "type": "template",
        "param_type": "input",
        "value": "DBRelease=${DBR}",
        "dataset": "ddo.000001.Atlas.Ideal.DBRelease.v220701",
        "attributes": "repeat,nosplit",
    },
    {"type": "constant", "value": "AMITag=p1462"},
    {
        "type": "template",
        "param_type": "output",
        "token": "ATLASDATADISK",
        "value": f"outputNTUP_EMBLLDNFile={outDatasetName}.${{SN}}.pool.root",
        "dataset": outDatasetName,
    },
]

jonStr = json.dumps(taskParamMap)

from pandajedi.jedicore.JediTaskBufferInterface import (  # noqa: E402
    JediTaskBufferInterface,
)

tbIF = JediTaskBufferInterface()
tbIF.setupInterface()
tbIF.insertTaskParams_JEDI(taskParamMap["vo"], taskParamMap["prodSourceLabel"], taskParamMap["userName"], taskParamMap["taskName"], jonStr)



import json

from pandacommon.pandalogger import logger_utils
from pandajedi.jedimsgprocessor.base_msg_processor import BaseMsgProcPlugin
from pandajedi.jedimsgprocessor.hpo_msg_processor import HPOMsgProcPlugin
from pandajedi.jedimsgprocessor.processing_msg_processor import ProcessingMsgProcPlugin
from pandajedi.jedimsgprocessor.tape_carousel_msg_processor import (
    TapeCarouselMsgProcPlugin,
)

# logger
base_logger = logger_utils.setup_logger(__name__.split(".")[-1])


# Atlas iDDS message processing plugin, a bridge connect to other idds related message processing plugins
class AtlasIddsMsgProcPlugin(BaseMsgProcPlugin):
    def initialize(self):
        BaseMsgProcPlugin.initialize(self)
        self.plugin_TapeCarousel = TapeCarouselMsgProcPlugin()
        self.plugin_HPO = HPOMsgProcPlugin()
        self.plugin_Processing = ProcessingMsgProcPlugin()
        # for each plugin
        for _plugin in [self.plugin_TapeCarousel, self.plugin_HPO, self.plugin_Processing]:
            # initialize each
            _plugin.initialize(in_collective=True)
            # use the same taskBuffer interface
            _plugin.tbIF = self.tbIF

    def process(self, msg_obj):
        # logger
        tmp_log = logger_utils.make_logger(base_logger, token=self.get_pid(), method_name="process")
        # start
        tmp_log.info("start")
        tmp_log.debug(f"sub_id={msg_obj.sub_id} ; msg_id={msg_obj.msg_id}")
        # parse json
        try:
            msg_dict = json.loads(msg_obj.data)
        except Exception as e:
            err_str = f"failed to parse message json {msg_obj.data} , skipped. {e.__class__.__name__} : {e}"
            tmp_log.error(err_str)
            raise
        # sanity check
        try:
            msg_type = msg_dict["msg_type"]
        except Exception as e:
            err_str = f"failed to parse message object dict {msg_dict} , skipped. {e.__class__.__name__} : {e}"
            tmp_log.error(err_str)
            raise
        # run different plugins according to message type
        try:
            if msg_type in ("file_stagein", "collection_stagein", "work_stagein"):
                tmp_log.debug("to tape_carousel")
                self.plugin_TapeCarousel.process(msg_obj, decoded_data=msg_dict)
            elif msg_type in ("file_hyperparameteropt", "collection_hyperparameteropt", "work_hyperparameteropt"):
                tmp_log.debug("to hpo")
                self.plugin_HPO.process(msg_obj, decoded_data=msg_dict)
            elif msg_type in ("file_processing", "collection_processing", "work_processing"):
                tmp_log.debug("to processing")
                self.plugin_Processing.process(msg_obj, decoded_data=msg_dict)
            else:
                # Asked by iDDS and message broker guys, JEDI needs to consume unknown types of messages and do nothing...
                warn_str = f"unknown msg_type : {msg_type}"
                tmp_log.warning(warn_str)
        except Exception:
            raise
        # done
        tmp_log.info("done")



import datetime
import random
import re
import shlex
import traceback

from pandajedi.jedicore import JediException
from pandaserver.dataservice import DataServiceUtils

from .TaskRefinerBase import TaskRefinerBase


# brokerage for ATLAS production
class AtlasProdTaskRefiner(TaskRefinerBase):
    # constructor
    def __init__(self, taskBufferIF, ddmIF):
        TaskRefinerBase.__init__(self, taskBufferIF, ddmIF)

    # extract common parameters
    def extractCommon(self, jediTaskID, taskParamMap, workQueueMapper, splitRule):
        tmpLog = self.tmpLog
        # set ddmBackEnd
        if "ddmBackEnd" not in taskParamMap:
            taskParamMap["ddmBackEnd"] = "rucio"
        # get number of unprocessed events for event service
        autoEsConversion = False
        if "esConvertible" in taskParamMap and taskParamMap["esConvertible"] is True:
            maxPrio = self.taskBufferIF.getConfigValue("taskrefiner", "AES_MAXTASKPRIORITY", "jedi", "atlas")
            minPrio = self.taskBufferIF.getConfigValue("taskrefiner", "AES_MINTASKPRIORITY", "jedi", "atlas")
            if maxPrio is not None and maxPrio < taskParamMap["taskPriority"]:
                pass
            elif minPrio is not None and minPrio > taskParamMap["taskPriority"]:
                pass
            else:
                # get threshold
                minNumEvents = self.taskBufferIF.getConfigValue("taskrefiner", "AES_EVENTPOOLSIZE", "jedi", "atlas")
                maxPending = self.taskBufferIF.getConfigValue("taskrefiner", "AES_MAXPENDING", "jedi", "atlas")
                nEvents, lastTaskTime, nPendingTasks = self.taskBufferIF.getNumUnprocessedEvents_JEDI(
                    taskParamMap["vo"], taskParamMap["prodSourceLabel"], {"eventService": 1}, {"gshare": "Validation"}
                )
                tmpStr = "check for ES "
                tmpStr += f"tot_num_unprocessed_events_AES={nEvents} target_num_events_AES={minNumEvents} last_AES_task_time={lastTaskTime} "
                tmpStr += f"num_pending_tasks_AES={nPendingTasks} max_pending_tasks_AES={maxPending} "
                tmpLog.info(tmpStr)
                # not chane many tasks at once
                if lastTaskTime is None or (lastTaskTime < datetime.datetime.now(datetime.timezone.utc).replace(tzinfo=None) - datetime.timedelta(minutes=5)):
                    if minNumEvents is not None and nEvents < minNumEvents and maxPending is not None and (maxPending is None or maxPending > nPendingTasks):
                        autoEsConversion = True
                        tmpLog.info("will be converted to AES unless it goes to pending")
        # add ES paramsters
        if ("esFraction" in taskParamMap and taskParamMap["esFraction"] > 0) or ("esConvertible" in taskParamMap and taskParamMap["esConvertible"] is True):
            tmpStr = "<PANDA_ES_ONLY>--eventService=True</PANDA_ES_ONLY>"
            taskParamMap["jobParameters"].append({"type": "constant", "value": tmpStr})
        if ("esFraction" in taskParamMap and taskParamMap["esFraction"] > 0) or autoEsConversion:
            if "nEventsPerWorker" not in taskParamMap and (("esFraction" in taskParamMap and taskParamMap["esFraction"] > random.random()) or autoEsConversion):
                taskParamMap["nEventsPerWorker"] = 1
                taskParamMap["registerEsFiles"] = True
                if "nEsConsumers" not in taskParamMap:
                    tmpVal = self.taskBufferIF.getConfigValue("taskrefiner", "AES_NESCONSUMERS", "jedi", "atlas")
                    if tmpVal is None:
                        tmpVal = 1
                    taskParamMap["nEsConsumers"] = tmpVal
                if "nSitesPerJob" not in taskParamMap:
                    tmpVal = self.taskBufferIF.getConfigValue("taskrefiner", "AES_NSITESPERJOB", "jedi", "atlas")
                    if tmpVal is not None:
                        taskParamMap["nSitesPerJob"] = tmpVal
                if "mergeEsOnOS" not in taskParamMap:
                    taskParamMap["mergeEsOnOS"] = True
                if "maxAttemptES" not in taskParamMap:
                    taskParamMap["maxAttemptES"] = 1
                if "maxAttemptEsJob" not in taskParamMap:
                    taskParamMap["maxAttemptEsJob"] = 0
                if "notDiscardEvents" not in taskParamMap:
                    taskParamMap["notDiscardEvents"] = True
                if "decAttOnFailedES" not in taskParamMap:
                    taskParamMap["decAttOnFailedES"] = True
                taskParamMap["coreCount"] = 0
                taskParamMap["resurrectConsumers"] = True
        TaskRefinerBase.extractCommon(self, jediTaskID, taskParamMap, workQueueMapper, splitRule)

    # main
    def doRefine(self, jediTaskID, taskParamMap):
        # make logger
        tmpLog = self.tmpLog
        tmpLog.debug(f"start taskType={self.taskSpec.taskType}")
        try:
            # basic refine
            self.doBasicRefine(taskParamMap)
            # set nosplit+repeat for DBR
            for datasetSpec in self.inSecDatasetSpecList:
                if DataServiceUtils.isDBR(datasetSpec.datasetName):
                    datasetSpec.attributes = "repeat,nosplit"
            # enable consistency check
            if self.taskSpec.parent_tid not in [None, self.taskSpec.jediTaskID]:
                for datasetSpec in self.inMasterDatasetSpec:
                    if datasetSpec.isMaster() and datasetSpec.type == "input":
                        datasetSpec.enableCheckConsistency()
            # append attempt number
            for tmpKey, tmpOutTemplateMapList in self.outputTemplateMap.items():
                for tmpOutTemplateMap in tmpOutTemplateMapList:
                    outFileTemplate = tmpOutTemplateMap["filenameTemplate"]
                    if re.search("\.\d+$", outFileTemplate) is None and not outFileTemplate.endswith(".panda.um"):
                        tmpOutTemplateMap["filenameTemplate"] = outFileTemplate + ".1"
            # extract input datatype
            datasetTypeListIn = []
            for datasetSpec in self.inMasterDatasetSpec + self.inSecDatasetSpecList:
                datasetType = DataServiceUtils.getDatasetType(datasetSpec.datasetName)
                if datasetType not in ["", None]:
                    datasetTypeListIn.append(datasetType)
            # extract datatype and set destination if nessesary
            datasetTypeList = []
            for datasetSpec in self.outDatasetSpecList:
                datasetType = DataServiceUtils.getDatasetType(datasetSpec.datasetName)
                if datasetType not in ["", None]:
                    datasetTypeList.append(datasetType)
            # set numThrottled to use the task throttling mechanism
            if "noThrottle" not in taskParamMap:
                self.taskSpec.numThrottled = 0
            # set to register datasets
            self.taskSpec.setToRegisterDatasets()
            # set transient to parent datasets
            if self.taskSpec.processingType in ["merge"] and self.taskSpec.parent_tid not in [None, self.taskSpec.jediTaskID]:
                # get parent
                tmpStat, parentTaskSpec = self.taskBufferIF.getTaskDatasetsWithID_JEDI(self.taskSpec.parent_tid, None, False)
                if tmpStat and parentTaskSpec is not None:
                    # set transient to parent datasets
                    metaData = {"transient": True}
                    for datasetSpec in parentTaskSpec.datasetSpecList:
                        if datasetSpec.type in ["log", "output"]:
                            datasetType = DataServiceUtils.getDatasetType(datasetSpec.datasetName)
                            if datasetType not in ["", None] and datasetType in datasetTypeList and datasetType in datasetTypeListIn:
                                tmpLog.info(
                                    "set metadata={0} to parent jediTaskID={1}:datasetID={2}:Name={3}".format(
                                        str(metaData), self.taskSpec.parent_tid, datasetSpec.datasetID, datasetSpec.datasetName
                                    )
                                )
                                for metadataName, metadaValue in metaData.items():
                                    self.ddmIF.getInterface(self.taskSpec.vo).setDatasetMetadata(datasetSpec.datasetName, metadataName, metadaValue)
            # input prestaging
            if self.taskSpec.inputPreStaging():
                # set first contents feed flag
                self.taskSpec.set_first_contents_feed(True)
        except JediException.UnknownDatasetError as e:
            tmpLog.debug(f"in doRefine. {str(e)}")
            raise e
        except Exception as e:
            tmpLog.error(f"doRefine failed with {str(e)} {traceback.format_exc()}")
            raise e
        tmpLog.debug("done")
        return self.SC_SUCCEEDED

    # insert string
    def insertString(self, paramName, tmpStr, origStr):
        items = shlex.split(origStr, posix=False)
        newStr = ""
        for item in items:
            if paramName not in item:
                newStr += item
            else:
                newStr += item[:-1]
                newStr += tmpStr
                newStr += item[-1]
            newStr += " "
        return newStr



import sys

from pandaserver.dataservice import DataServiceUtils
from pandaserver.taskbuffer import EventServiceUtils

from . import AtlasPostProcessorUtils
from .PostProcessorBase import PostProcessorBase


# post processor for ATLAS production
class AtlasProdPostProcessor(PostProcessorBase):
    # constructor
    def __init__(self, taskBufferIF, ddmIF):
        PostProcessorBase.__init__(self, taskBufferIF, ddmIF)

    # main
    def doPostProcess(self, taskSpec, tmpLog):
        # pre-check
        try:
            tmpStat = self.doPreCheck(taskSpec, tmpLog)
            if tmpStat:
                return self.SC_SUCCEEDED
        except Exception:
            errtype, errvalue = sys.exc_info()[:2]
            tmpLog.error(f"doPreCheck failed with {errtype.__name__}:{errvalue}")
            return self.SC_FATAL
        # get DDM I/F
        ddmIF = self.ddmIF.getInterface(taskSpec.vo)
        # loop over all datasets
        for datasetSpec in taskSpec.datasetSpecList:
            # skip pseudo output datasets
            if datasetSpec.type in ["output"] and datasetSpec.isPseudo():
                continue
            try:
                # remove wrong files
                if datasetSpec.type in ["output"]:
                    # get successful files
                    okFiles = self.taskBufferIF.getSuccessfulFiles_JEDI(datasetSpec.jediTaskID, datasetSpec.datasetID)
                    if okFiles is None:
                        tmpLog.warning(f"failed to get successful files for {datasetSpec.datasetName}")
                        return self.SC_FAILED
                    # get files in dataset
                    ddmFiles = ddmIF.getFilesInDataset(datasetSpec.datasetName, skipDuplicate=False, ignoreUnknown=True)
                    tmpLog.debug(
                        f"datasetID={datasetSpec.datasetID}:Name={datasetSpec.datasetName} has {len(okFiles)} files in DB, {len(ddmFiles)} files in DDM"
                    )
                    # check all files
                    toDelete = []
                    for tmpGUID, attMap in ddmFiles.items():
                        if attMap["lfn"] not in okFiles:
                            did = {"scope": attMap["scope"], "name": attMap["lfn"]}
                            toDelete.append(did)
                            tmpLog.debug(f"delete {attMap['lfn']} from {datasetSpec.datasetName}")
                    # delete
                    if toDelete != []:
                        ddmIF.deleteFilesFromDataset(datasetSpec.datasetName, toDelete)
            except Exception:
                errtype, errvalue = sys.exc_info()[:2]
                tmpLog.warning(f"failed to remove wrong files with {errtype.__name__}:{errvalue}")
                return self.SC_FAILED
            try:
                # freeze output and log datasets
                if datasetSpec.type in ["output", "log", "trn_log"]:
                    tmpLog.info(f"freezing datasetID={datasetSpec.datasetID}:Name={datasetSpec.datasetName}")
                    ddmIF.freezeDataset(datasetSpec.datasetName, ignoreUnknown=True)
            except Exception:
                errtype, errvalue = sys.exc_info()[:2]
                tmpLog.warning(f"failed to freeze datasets with {errtype.__name__}:{errvalue}")
                return self.SC_FAILED
            try:
                # delete transient datasets
                if datasetSpec.type in ["trn_output"]:
                    tmpLog.debug(f"deleting datasetID={datasetSpec.datasetID}:Name={datasetSpec.datasetName}")
                    retStr = ddmIF.deleteDataset(datasetSpec.datasetName, False, ignoreUnknown=True)
                    tmpLog.info(retStr)
            except Exception:
                errtype, errvalue = sys.exc_info()[:2]
                tmpLog.warning(f"failed to delete datasets with {errtype.__name__}:{errvalue}")
        # check duplication
        if self.getFinalTaskStatus(taskSpec) in ["finished", "done"] and taskSpec.gshare != "Test":
            nDup = self.taskBufferIF.checkDuplication_JEDI(taskSpec.jediTaskID)
            tmpLog.debug(f"checked duplication with {nDup}")
            if nDup is not None and nDup > 0:
                errStr = f"paused since {nDup} duplication found"
                taskSpec.oldStatus = self.getFinalTaskStatus(taskSpec)
                taskSpec.status = "paused"
                taskSpec.setErrDiag(errStr)
                tmpLog.debug(errStr)
        # delete ES datasets
        if taskSpec.registerEsFiles():
            try:
                targetName = EventServiceUtils.getEsDatasetName(taskSpec.jediTaskID)
                tmpLog.debug(f"deleting ES dataset name={targetName}")
                retStr = ddmIF.deleteDataset(targetName, False, ignoreUnknown=True)
                tmpLog.debug(retStr)
            except Exception:
                errtype, errvalue = sys.exc_info()[:2]
                tmpLog.warning(f"failed to delete ES dataset with {errtype.__name__}:{errvalue}")
        try:
            AtlasPostProcessorUtils.send_notification(self.taskBufferIF, ddmIF, taskSpec, tmpLog)
        except Exception as e:
            tmpLog.error(f"failed to talk to external system with {str(e)}")
            return self.SC_FAILED
        try:
            self.doBasicPostProcess(taskSpec, tmpLog)
        except Exception:
            errtype, errvalue = sys.exc_info()[:2]
            tmpLog.error(f"doBasicPostProcess failed with {errtype.__name__}:{errvalue}")
            return self.SC_FATAL
        return self.SC_SUCCEEDED

    # final procedure
    def doFinalProcedure(self, taskSpec, tmpLog):
        tmpLog.info(f"final procedure for status={taskSpec.status} processingType={taskSpec.processingType}")
        if taskSpec.status in ["done", "finished"] or (taskSpec.status == "paused" and taskSpec.oldStatus in ["done", "finished"]):
            trnLifeTime = 14 * 24 * 60 * 60
            trnLifeTimeMerge = 40 * 24 * 60 * 60
            ddmIF = self.ddmIF.getInterface(taskSpec.vo)
            # set lifetime to transient datasets
            metaData = {"lifetime": trnLifeTime}
            datasetTypeListI = set()
            datasetTypeListO = set()
            for datasetSpec in taskSpec.datasetSpecList:
                if datasetSpec.type in ["log", "output"]:
                    if datasetSpec.getTransient() is True:
                        tmpLog.debug(f"set metadata={str(metaData)} to datasetID={datasetSpec.datasetID}:Name={datasetSpec.datasetName}")
                        for metadataName, metadaValue in metaData.items():
                            ddmIF.setDatasetMetadata(datasetSpec.datasetName, metadataName, metadaValue)
                # collect dataset types
                datasetType = DataServiceUtils.getDatasetType(datasetSpec.datasetName)
                if datasetType not in ["", None]:
                    if datasetSpec.type == "input":
                        datasetTypeListI.add(datasetType)
                    elif datasetSpec.type == "output":
                        datasetTypeListO.add(datasetType)
            # set lifetime to parent transient datasets
            if taskSpec.processingType in ["merge"] and (
                taskSpec.status == "done"
                or (taskSpec.status == "finished" and self.getFinalTaskStatus(taskSpec, checkParent=False) == "done")
                or (
                    taskSpec.status == "paused"
                    and (taskSpec.oldStatus == "done" or (taskSpec.oldStatus == "finished" and self.getFinalTaskStatus(taskSpec, checkParent=False) == "done"))
                )
            ):
                # get parent task
                if taskSpec.parent_tid not in [None, taskSpec.jediTaskID]:
                    # get parent
                    tmpStat, parentTaskSpec = self.taskBufferIF.getTaskDatasetsWithID_JEDI(taskSpec.parent_tid, None, False)
                    if tmpStat and parentTaskSpec is not None:
                        # set lifetime to parent datasets if they are transient
                        for datasetSpec in parentTaskSpec.datasetSpecList:
                            if datasetSpec.type in ["output"]:
                                # check dataset type
                                datasetType = DataServiceUtils.getDatasetType(datasetSpec.datasetName)
                                if datasetType not in datasetTypeListI or datasetType not in datasetTypeListO:
                                    continue
                                metaData = {"lifetime": trnLifeTimeMerge}
                                tmpMetadata = ddmIF.getDatasetMetaData(datasetSpec.datasetName)
                                if tmpMetadata["transient"] is True:
                                    tmpLog.debug(
                                        "set metadata={0} to parent jediTaskID={1}:datasetID={2}:Name={3}".format(
                                            str(metaData), taskSpec.parent_tid, datasetSpec.datasetID, datasetSpec.datasetName
                                        )
                                    )
                                    for metadataName, metadaValue in metaData.items():
                                        ddmIF.setDatasetMetadata(datasetSpec.datasetName, metadataName, metadaValue)
        # set lifetime to failed datasets
        if taskSpec.status in ["failed", "broken", "aborted"]:
            trnLifeTime = 30 * 24 * 60 * 60
            ddmIF = self.ddmIF.getInterface(taskSpec.vo)
            # only log datasets
            metaData = {"lifetime": trnLifeTime}
            for datasetSpec in taskSpec.datasetSpecList:
                if datasetSpec.type in ["log"]:
                    tmpLog.debug(f"set metadata={str(metaData)} to failed datasetID={datasetSpec.datasetID}:Name={datasetSpec.datasetName}")
                    for metadataName, metadaValue in metaData.items():
                        ddmIF.setDatasetMetadata(datasetSpec.datasetName, metadataName, metadaValue)
        return self.SC_SUCCEEDED



import json

from pandacommon.pandalogger import logger_utils
from pandajedi.jediconfig import jedi_config

# from pandajedi.jedicore.FactoryBase import FactoryBase
# from pandajedi.jedicore.JediCoreUtils import convert_config_params, parse_init_params
from pandajedi.jedicore.ThreadUtils import ListWithLock, ThreadPool
from pandajedi.jediddm.DDMInterface import DDMInterface
from pandajedi.jedimsgprocessor.base_msg_processor import BaseMsgProcPlugin
from pandajedi.jediorder.JobGenerator import JobGeneratorThread
from pandajedi.jediorder.TaskSetupper import TaskSetupper

# logger
base_logger = logger_utils.setup_logger(__name__.split(".")[-1])


# Jedi Job Generator message processor plugin
class JediJobGeneratorMsgProcPlugin(BaseMsgProcPlugin):
    """
    Message-driven Post-Processor
    """

    def initialize(self):
        BaseMsgProcPlugin.initialize(self)
        # DDM interface
        self.ddmIF = DDMInterface()
        self.ddmIF.setupInterface()
        # get SiteMapper
        # siteMapper = self.tbIF.getSiteMapper()
        # get work queue mapper
        # workQueueMapper = self.tbIF.getWorkQueueMap()
        # get TaskSetupper
        # taskSetupper = TaskSetupper(self.vos, self.prodSourceLabels)
        # taskSetupper.initializeMods(self.tbIF, self.ddmIF)
        self.pid = self.get_pid()

    def process(self, msg_obj):
        # logger
        tmp_log = logger_utils.make_logger(base_logger, token=self.get_pid(), method_name="process")
        # start
        tmp_log.info("start")
        tmp_log.debug(f"sub_id={msg_obj.sub_id} ; msg_id={msg_obj.msg_id}")
        # parse json
        try:
            msg_dict = json.loads(msg_obj.data)
        except Exception as e:
            err_str = f"failed to parse message json {msg_obj.data} , skipped. {e.__class__.__name__} : {e}"
            tmp_log.error(err_str)
            raise
        # sanity check
        try:
            msg_type = msg_dict["msg_type"]
        except Exception as e:
            err_str = f"failed to parse message object dict {msg_dict} , skipped. {e.__class__.__name__} : {e}"
            tmp_log.error(err_str)
            raise
        if msg_type != "jedi_job_generator":
            # FIXME
            err_str = f"got unknown msg_type {msg_type} , skipped "
            tmp_log.error(err_str)
            raise
        # run
        try:
            # get task to generate jobs
            task_id = int(msg_dict["taskid"])
            _, taskSpec = self.tbIF.getTaskWithID_JEDI(task_id)
            if not taskSpec:
                tmp_log.debug(f"unknown task {task_id}")
            else:
                # get WQ
                vo = taskSpec.vo
                prodSourceLabel = taskSpec.prodSourceLabel
                workQueue = self.tbIF.getWorkQueueMap().getQueueWithIDGshare(taskSpec.workQueue_ID, taskSpec.gshare)
                # get inputs
                tmpList = self.tbIF.getTasksToBeProcessed_JEDI(self.pid, None, workQueue, None, None, nFiles=1000, target_tasks=[task_id])
                if tmpList:
                    inputList = ListWithLock(tmpList)
                    # create thread
                    threadPool = ThreadPool()
                    siteMapper = self.tbIF.getSiteMapper()
                    taskSetupper = TaskSetupper(vo, prodSourceLabel)
                    taskSetupper.initializeMods(self.tbIF, self.ddmIF)
                    gen_thr = JobGeneratorThread(
                        inputList, threadPool, self.tbIF, self.ddmIF, siteMapper, True, taskSetupper, self.pid, workQueue, "pjmsg", None, None, None, False
                    )
                    gen_thr.start()
                    gen_thr.join()
                    tmp_log.info(f"generated jobs for task {task_id}")
                else:
                    tmp_log.debug(f"task {task_id} is not considered to be processed; skipped ")
        except Exception as e:
            err_str = f"failed to run, skipped. {e.__class__.__name__} : {e}"
            tmp_log.error(err_str)
            raise
        # done
        tmp_log.info("done")



# logger
from pandacommon.pandalogger.PandaLogger import PandaLogger
from pandajedi.jedicore.MsgWrapper import MsgWrapper

from .DDMClientBase import DDMClientBase

logger = PandaLogger().getLogger(__name__.split(".")[-1])


# class to access to general DDM
class GenDDMClient(DDMClientBase):
    # constructor
    def __init__(self, con):
        # initialize base class
        DDMClientBase.__init__(self, con)

    # get dataset metadata
    def getDatasetMetaData(self, datasetName, ignore_missing=False):
        return self.SC_SUCCEEDED, {"state": "closed"}



import datetime
import re
import time

from pandacommon.pandalogger.PandaLogger import PandaLogger
from pandajedi.jediconfig import jedi_config
from pandaserver.userinterface import Client


class MsgWrapper:
    def __init__(self, logger, token=None, lineLimit=500, monToken=None):
        self.logger = logger
        # use timestamp as token if undefined
        if token is None:
            self.token = f"<{datetime.datetime.now(datetime.timezone.utc).replace(tzinfo=None).isoformat('/')}>"
        else:
            self.token = token
        # token for http logger
        if monToken is None:
            self.monToken = self.token
        else:
            self.monToken = monToken
        # remove <> for django
        try:
            self.monToken = re.sub("<(?P<name>[^>]+)>", "\g<name>", self.monToken)
        except Exception:
            pass
        # message buffer
        self.msgBuffer = []
        self.bareMsg = []
        self.lineLimit = lineLimit

    def keepMsg(self, msg):
        # keep max message depth
        if len(self.msgBuffer) > self.lineLimit:
            self.msgBuffer.pop(0)
            self.bareMsg.pop(0)
        timeNow = datetime.datetime.now(datetime.timezone.utc).replace(tzinfo=None)
        self.msgBuffer.append(f"{timeNow.isoformat(' ')} : {msg}")
        self.bareMsg.append(msg)

    def info(self, msg):
        msg = str(msg)
        self.logger.info(self.token + " " + msg)
        self.keepMsg(msg)

    def debug(self, msg):
        msg = str(msg)
        self.logger.debug(self.token + " " + msg)
        self.keepMsg(msg)

    def error(self, msg):
        msg = str(msg)
        self.logger.error(self.token + " " + msg)
        self.keepMsg(msg)

    def warning(self, msg):
        msg = str(msg)
        self.logger.warning(self.token + " " + msg)
        self.keepMsg(msg)

    def dumpToString(self):
        strMsg = ""
        for msg in self.msgBuffer:
            strMsg += msg
            strMsg += "\n"
        return strMsg

    def uploadLog(self, id):
        strMsg = self.dumpToString()
        s, o = Client.uploadLog(strMsg, id)
        if s != 0:
            return f"failed to upload log with {s}."
        if o.startswith("http"):
            return f"<a href=\"{o}\">log</a> : {'. '.join(self.bareMsg[-2:])}."
        return o

    # send message to logger
    def sendMsg(self, message, msgType, msgLevel="info", escapeChar=False):
        try:
            # get logger
            tmpPandaLogger = PandaLogger()
            # lock HTTP handler
            tmpPandaLogger.lock()
            tmpPandaLogger.setParams({"Type": msgType})
            # get logger
            tmpLogger = tmpPandaLogger.getHttpLogger(jedi_config.master.loggername)
            # escape special characters
            if escapeChar:
                message = message.replace("<", "&lt;")
                message = message.replace(">", "&gt;")
            # add message
            message = self.monToken + " " + message
            if msgLevel == "error":
                tmpLogger.error(message)
            elif msgLevel == "warning":
                tmpLogger.warning(message)
            elif msgLevel == "info":
                tmpLogger.info(message)
            else:
                tmpLogger.debug(message)
            # release HTTP handler
            tmpPandaLogger.release()
        except Exception:
            pass



import re

from pandajedi.jedicore.JediTaskSpec import JediTaskSpec

from .TaskRefinerBase import TaskRefinerBase


# refiner for general purpose
class GenTaskRefiner(TaskRefinerBase):
    # constructor
    def __init__(self, taskBufferIF, ddmIF):
        TaskRefinerBase.__init__(self, taskBufferIF, ddmIF)

    # extract common parameters
    def extractCommon(self, jediTaskID, taskParamMap, workQueueMapper, splitRule):
        if "cloud" not in taskParamMap and "workingGroup" in taskParamMap:
            taskParamMap["cloud"] = taskParamMap["workingGroup"]
        if "transPath" not in taskParamMap:
            taskParamMap["transPath"] = "http://pandaserver-doma.cern.ch:25080/trf/user/runGen-00-00-02"
        # set sourceURL
        try:
            if "sourceURL" in taskParamMap:
                for tmpItem in taskParamMap["jobParameters"]:
                    if "value" in tmpItem:
                        tmpItem["value"] = re.sub("\$\{SURL\}", taskParamMap["sourceURL"], tmpItem["value"])
        except Exception:
            pass
        # min ram count
        if "ramCount" not in taskParamMap:
            taskParamMap["ramCount"] = 2000
            taskParamMap["ramUnit"] = "MBPerCore"
        # push status changes
        if "pushStatusChanges" not in taskParamMap:
            taskParamMap["pushStatusChanges"] = True
        # fine-grained
        if "fineGrainedProc" in taskParamMap:
            taskParamMap["notDiscardEvents"] = True
            taskParamMap["maxAttemptEsJob"] = 0
            taskParamMap["maxAttemptES"] = 1
        # message driven
        if "messageDriven" not in taskParamMap:
            taskParamMap["messageDriven"] = True
        # use cloud as VO
        taskParamMap["cloudAsVO"] = True
        # update task parameters
        self.updatedTaskParams = taskParamMap
        # call base method
        TaskRefinerBase.extractCommon(self, jediTaskID, taskParamMap, workQueueMapper, splitRule)

    # main
    def doRefine(self, jediTaskID, taskParamMap):
        # normal refine
        self.doBasicRefine(taskParamMap)
        # get DDM I/F to check
        if self.ddmIF.getInterface(self.taskSpec.vo, self.taskSpec.cloud):
            # use template dataset
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["instantiateTmpl"])
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["instantiateTmplSite"])
            for datasetSpec in self.outDatasetSpecList:
                datasetSpec.type = f"tmpl_{datasetSpec.type}"
        return self.SC_SUCCEEDED



import json

from pandacommon.pandalogger import logger_utils
from pandajedi.jedimsgprocessor.base_msg_processor import BaseMsgProcPlugin

# logger
base_logger = logger_utils.setup_logger(__name__.split(".")[-1])


# Hyper-Parameter-Optimization message processing plugin
class HPOMsgProcPlugin(BaseMsgProcPlugin):
    def process(self, msg_obj, decoded_data=None):
        # logger
        tmp_log = logger_utils.make_logger(base_logger, token=self.get_pid(), method_name="process")
        # start
        tmp_log.info("start")
        tmp_log.debug(f"sub_id={msg_obj.sub_id} ; msg_id={msg_obj.msg_id}")
        # parse
        if decoded_data is None:
            # json decode
            try:
                msg_dict = json.loads(msg_obj.data)
            except Exception as e:
                err_str = f"failed to parse message json {msg_obj.data} , skipped. {e.__class__.__name__} : {e}"
                tmp_log.error(err_str)
                raise
        else:
            msg_dict = decoded_data
        # sanity check
        try:
            msg_type = msg_dict["msg_type"]
            jeditaskid = int(msg_dict["workload_id"])
            if msg_type == "file_hyperparameteropt":
                target_list = msg_dict["files"]
            elif msg_type == "collection_hyperparameteropt":
                # to finish the task
                pass
            elif msg_type == "work_hyperparameteropt":
                pass
            else:
                raise ValueError(f"invalid msg_type value: {msg_type}")
        except Exception as e:
            err_str = f"failed to parse message object dict {msg_dict} , skipped. {e.__class__.__name__} : {e}"
            tmp_log.error(err_str)
            raise
        # run
        if msg_type == "file_hyperparameteropt":
            # insert HPO events
            try:
                # event ids from the targets
                event_id_list = []
                for target in target_list:
                    if target["status"] != "New":
                        continue
                    model_id = None
                    try:
                        path = json.loads(target["path"])
                        if isinstance(path[0], (list, tuple)):
                            model_id = path[0][0]
                    except Exception:
                        pass
                    event_id_list.append((target["name"], model_id))
                if event_id_list:
                    n_events = len(event_id_list)
                    # insert events
                    res = self.tbIF.insertHpoEventAboutIdds_JEDI(jedi_task_id=jeditaskid, event_id_list=event_id_list)
                    # check if ok
                    if res:
                        tmp_log.debug(f"jeditaskid={jeditaskid}, inserted {n_events} events: {event_id_list}")
                    else:
                        tmp_log.warning(f"jeditaskid={jeditaskid}, failed to insert events: {event_id_list}")
            except Exception as e:
                err_str = f"failed to parse message object, skipped. {e.__class__.__name__} : {e}"
                tmp_log.error(err_str)
                raise
        elif msg_type == "collection_hyperparameteropt":
            # finish the task
            try:
                # send finish command
                retVal, retStr = self.tbIF.sendCommandTaskPanda(jeditaskid, "iDDS. HPO task finished", True, "finish", comQualifier="soft")
                # check if ok
                if retVal:
                    tmp_log.debug(f"jeditaskid={jeditaskid}, finished the task")
                else:
                    tmp_log.warning(f"jeditaskid={jeditaskid}, failed finish the task: {retStr}")
            except Exception as e:
                err_str = f"failed to parse message object, skipped. {e.__class__.__name__} : {e}"
                tmp_log.error(err_str)
                raise
        else:
            # do nothing
            tmp_log.debug(f"jeditaskid={jeditaskid}, msg_type={msg_type}, did nothing")
        # done
        tmp_log.info("done")



import sys
import uuid

from userinterface import Client

inFileList = ["file1"]

logDatasetName = f"panda.jeditest.log.{uuid.uuid4()}"

taskParamMap = {}

taskParamMap["nFilesPerJob"] = 1
taskParamMap["nFiles"] = len(inFileList)
# taskParamMap['nEventsPerInputFile']  = 10000
# taskParamMap['nEventsPerJob'] = 10000
# taskParamMap['nEvents'] = 25000
taskParamMap["noInput"] = True
taskParamMap["pfnList"] = inFileList
# taskParamMap['mergeOutput'] = True
taskParamMap["taskName"] = str(uuid.uuid4())
taskParamMap["userName"] = "someone"
taskParamMap["vo"] = "wlcg"
taskParamMap["taskPriority"] = 900
# taskParamMap['reqID'] = reqIdx
taskParamMap["architecture"] = ""
taskParamMap["transUses"] = ""
taskParamMap["transHome"] = None
# taskParamMap['transPath'] = 'https://atlpan.web.cern.ch/atlpan/bash-c'
taskParamMap["transPath"] = "https://atlpan.web.cern.ch/atlpan/bash-c-enc"
taskParamMap["encJobParams"] = True
taskParamMap["processingType"] = "step1"
taskParamMap["prodSourceLabel"] = "test"
taskParamMap["taskType"] = "test"
taskParamMap["workingGroup"] = "lsst"
# taskParamMap['coreCount'] = 1
# taskParamMap['walltime'] = 1
taskParamMap["cloud"] = "LSST"
taskParamMap["site"] = "DOMA_LSST_GOOGLE_TEST"
# taskParamMap['site'] = 'DOMA_LSST_SLAC_TEST'

# taskParamMap['ramCount'] = 1000
taskParamMap["ramUnit"] = "MB"

"""
taskParamMap['log'] = {'dataset': logDatasetName,
                       'type':'template',
                       'param_type':'log',
                       'token':'local',
                       'destination':'local',
                       'value':'{0}.${{SN}}.log.tgz'.format(logDatasetName)}
"""

taskParamMap["jobParameters"] = [
    {
        "type": "constant",
        "value": "echo aaa; ls; echo",
    },
    {
        "type": "constant",
        "value": "\"'${IN/L}'\"",
    },
]

print(Client.insertTaskParams(taskParamMap))



"""
cache specification for JEDI

"""


class JediCacheSpec(object):
    # attributes
    attributes = (
        "main_key",
        "sub_key",
        "data",
        "last_update",
    )

    # constructor
    def __init__(self):
        # install attributes
        for attr in self.attributes:
            object.__setattr__(self, attr, None)

    # return map of values
    def valuesMap(self):
        ret = {}
        for attr in self.attributes:
            val = getattr(self, attr)
            ret[f":{attr}"] = val
        return ret

    # pack tuple into JediCacheSpec
    def pack(self, values):
        for i, attr in enumerate(self.attributes):
            val = values[i]
            object.__setattr__(self, attr, val)

    # return column names for INSERT
    @classmethod
    def columnNames(cls, prefix=None):
        ret = ""
        if prefix is None:
            ret = ",".join(cls.attributes)
        else:
            ret = ",".join([f"{prefix}.{attr}" for attr in cls.attributes])
        return ret

    # return expression of bind variables for INSERT
    @classmethod
    def bindValuesExpression(cls):
        values_str = ",".join([f":{attr}" for attr in cls.attributes])
        ret = f"VALUES({values_str})"
        return ret

    # return an expression of bind variables for UPDATE
    @classmethod
    def bindUpdateChangesExpression(cls):
        ret = ",".join(["{0}=:{0}".format(attr) for attr in cls.attributes])
        return ret



import json

from pandacommon.pandalogger import logger_utils
from pandajedi.jediconfig import jedi_config
from pandajedi.jedicore.FactoryBase import FactoryBase
from pandajedi.jedicore.JediCoreUtils import convert_config_params, parse_init_params
from pandajedi.jediddm.DDMInterface import DDMInterface
from pandajedi.jedimsgprocessor.base_msg_processor import BaseMsgProcPlugin
from pandajedi.jediorder.PostProcessor import PostProcessorThread

# logger
base_logger = logger_utils.setup_logger(__name__.split(".")[-1])


# Jedi Post-Processor message processor plugin
class JediPostProcessorMsgProcPlugin(BaseMsgProcPlugin):
    """
    Message-driven Post-Processor
    """

    def initialize(self):
        BaseMsgProcPlugin.initialize(self)
        # DDM interface
        ddmIF = DDMInterface()
        ddmIF.setupInterface()
        # factory bases and post processor thread object for all vos and prodsourcelabels
        self.post_processor_thread_dict = dict()
        for itemStr in jedi_config.postprocessor.procConfig.split(";"):
            items = convert_config_params(itemStr)
            vos = parse_init_params(items[0])
            prodsourcelabels = parse_init_params(items[1])
            tmp_factory_base_obj = FactoryBase(vos=vos, sourceLabels=prodsourcelabels, logger=base_logger, modConfig=jedi_config.postprocessor.modConfig)
            tmp_post_processor_thread_obj = PostProcessorThread(
                taskList=None, threadPool=None, taskbufferIF=self.tbIF, ddmIF=ddmIF, implFactory=tmp_factory_base_obj
            )
            for vo in vos:
                for prodsourcelabel in prodsourcelabels:
                    self.post_processor_thread_dict[(vo, prodsourcelabel)] = tmp_post_processor_thread_obj

    def process(self, msg_obj):
        # logger
        tmp_log = logger_utils.make_logger(base_logger, token=self.get_pid(), method_name="process")
        # start
        tmp_log.info("start")
        tmp_log.debug(f"sub_id={msg_obj.sub_id} ; msg_id={msg_obj.msg_id}")
        # parse json
        try:
            msg_dict = json.loads(msg_obj.data)
        except Exception as e:
            err_str = f"failed to parse message json {msg_obj.data} , skipped. {e.__class__.__name__} : {e}"
            tmp_log.error(err_str)
            raise
        # sanity check
        try:
            msg_type = msg_dict["msg_type"]
        except Exception as e:
            err_str = f"failed to parse message object dict {msg_dict} , skipped. {e.__class__.__name__} : {e}"
            tmp_log.error(err_str)
            raise
        if msg_type != "jedi_post_processor":
            # FIXME
            err_str = f"got unknown msg_type {msg_type} , skipped "
            tmp_log.error(err_str)
            raise
        # run
        try:
            task_id = msg_dict["taskid"]
            vo = msg_dict["task_vo"]
            prodsourcelabel = msg_dict["task_prodsourcelabel"]
            ret_list = self.tbIF.prepareTasksToBeFinished_JEDI(vo, prodSourceLabel, jedi_config.postprocessor.nTasks, self.get_pid())
            task_list = self.tbIF.getTasksToBeFinished_JEDI(vo, prodSourceLabel, self.get_pid(), jedi_config.postprocessor.nTasks, target_tasks=ret_list)
            if task_list and task_id in [task_spec.jediTaskID for task_spec in task_list]:
                tmp_post_processor_thread_obj = self.post_processor_thread_dict[(vo, prodsourcelabel)]
                tmp_post_processor_thread_obj.post_process_tasks(task_list)
                tmp_log.info(f"post processed tasks {task_list} including {task_id}")
            else:
                tmp_log.debug(f"did not get task {task_id}; skip ")
        except Exception as e:
            err_str = f"failed to run, skipped. {e.__class__.__name__} : {e}"
            tmp_log.error(err_str)
            raise
        # done
        tmp_log.info("done")



import sys

from .MsgWrapper import MsgWrapper

_factoryModuleName = __name__.split(".")[-1]


# base class for factory
class FactoryBase:
    # constructor
    def __init__(self, vos, sourceLabels, logger, modConfig):
        if isinstance(vos, list):
            self.vos = vos
        else:
            try:
                self.vos = vos.split("|")
            except Exception:
                self.vos = [vos]
        if isinstance(sourceLabels, list):
            self.sourceLabels = sourceLabels
        else:
            try:
                self.sourceLabels = sourceLabels.split("|")
            except Exception:
                self.sourceLabels = [sourceLabels]
        self.modConfig = modConfig
        self.logger = MsgWrapper(logger, _factoryModuleName)
        self.implMap = {}
        self.className = None
        self.classMap = {}

    # initialize all modules
    def initializeMods(self, *args):
        # parse config
        for configStr in self.modConfig.split(","):
            configStr = configStr.strip()
            items = configStr.split(":")
            # check format
            try:
                vos = items[0].split("|")
                sourceLabels = items[1].split("|")
                moduleName = items[2]
                className = items[3]
                try:
                    subTypes = items[4].split("|")
                except Exception:
                    subTypes = ["any"]
            except Exception:
                self.logger.error(f"wrong config definition : {configStr}")
                continue
            # loop over all VOs
            for vo in vos:
                # loop over all labels
                for sourceLabel in sourceLabels:
                    # check vo and sourceLabel if specified
                    if vo not in ["", "any"] and vo not in self.vos and None not in self.vos and "any" not in self.vos:
                        continue
                    if (
                        sourceLabel not in ["", "any"]
                        and sourceLabel not in self.sourceLabels
                        and None not in self.sourceLabels
                        and "any" not in self.sourceLabels
                    ):
                        continue
                    # loop over all sub types
                    for subType in subTypes:
                        # import
                        try:
                            # import module
                            self.logger.info(f"vo={vo} label={sourceLabel} subtype={subType}")
                            self.logger.info(f"importing {moduleName}")
                            mod = __import__(moduleName)
                            for subModuleName in moduleName.split(".")[1:]:
                                mod = getattr(mod, subModuleName)
                            # get class
                            self.logger.info(f"getting class {className}")
                            cls = getattr(mod, className)
                            # instantiate
                            self.logger.info("instantiating")
                            impl = cls(*args)
                            # set vo
                            impl.vo = vo
                            impl.prodSourceLabel = sourceLabel
                            # append
                            if vo not in self.implMap:
                                self.implMap[vo] = {}
                                self.classMap[vo] = {}
                            if sourceLabel not in self.implMap[vo]:
                                self.implMap[vo][sourceLabel] = {}
                                self.classMap[vo][sourceLabel] = {}
                            self.implMap[vo][sourceLabel][subType] = impl
                            self.classMap[vo][sourceLabel][subType] = cls
                            self.logger.info(f"{cls} is ready for {vo}:{sourceLabel}:{subType}")
                        except Exception:
                            errtype, errvalue = sys.exc_info()[:2]
                            self.logger.error(
                                "failed to import {mn}.{cn} for vo={vo} label={lb} subtype={st} due to {et} {ev}".format(
                                    et=errtype.__name__, ev=errvalue, st=subType, vo=vo, lb=sourceLabel, cn=className, mn=moduleName
                                )
                            )
                            raise ImportError(f"failed to import {moduleName}.{className}")
        # return
        return True

    # get implementation for vo and sourceLabel. Only work with initializeMods()
    def getImpl(self, vo, sourceLabel, subType="any", doRefresh=True):
        # check VO
        if vo in self.implMap:
            # match VO
            voImplMap = self.implMap[vo]
        elif "any" in self.implMap:
            # catch all
            voImplMap = self.implMap["any"]
        else:
            return None
        # check sourceLabel
        if sourceLabel in voImplMap:
            # match sourceLabel
            srcImplMap = voImplMap[sourceLabel]
        elif "any" in voImplMap:
            # catch all
            srcImplMap = voImplMap["any"]
        else:
            return None
        # check subType
        if subType in srcImplMap:
            # match subType
            tmpImpl = srcImplMap[subType]
            if doRefresh:
                tmpImpl.refresh()
            return tmpImpl
        elif "any" in srcImplMap:
            # catch all
            tmpImpl = srcImplMap["any"]
            if doRefresh:
                tmpImpl.refresh()
            return tmpImpl
        else:
            return None

    # instantiate implementation for vo and sourceLabel. Only work with initializeMods()
    def instantiateImpl(self, vo, sourceLabel, subType, *args):
        # check VO
        if vo in self.classMap:
            # match VO
            voImplMap = self.classMap[vo]
        elif "any" in self.classMap:
            # catch all
            voImplMap = self.classMap["any"]
        else:
            return None
        # check sourceLabel
        if sourceLabel in voImplMap:
            # match sourceLabel
            srcImplMap = voImplMap[sourceLabel]
        elif "any" in voImplMap:
            # catch all
            srcImplMap = voImplMap["any"]
        else:
            return None
        # check subType
        if subType in srcImplMap:
            # match subType
            impl = srcImplMap[subType](*args)
            impl.vo = vo
            impl.prodSourceLabel = sourceLabel
            return impl
        elif "any" in srcImplMap:
            # catch all
            impl = srcImplMap["any"](*args)
            impl.vo = vo
            impl.prodSourceLabel = sourceLabel
            return impl
        else:
            return None

    # get class name of impl
    def getClassName(self, vo=None, sourceLabel=None):
        impl = self.getImpl(vo, sourceLabel, doRefresh=False)
        if impl is None:
            return None
        return impl.__class__.__name__



from .PostProcessorBase import PostProcessorBase


# post processor for general purpose
class GenPostProcessor(PostProcessorBase):
    # constructor
    def __init__(self, taskBufferIF, ddmIF):
        PostProcessorBase.__init__(self, taskBufferIF, ddmIF)
        self.failOnZeroOkFile = True

    # main
    def doPostProcess(self, taskSpec, tmpLog):
        try:
            # get DDM I/F
            ddmIF = self.ddmIF.getInterface(taskSpec.vo, taskSpec.cloud)
            # skip if DDM I/F is inactive
            if not ddmIF:
                tmpLog.info("skip due to inactive DDM I/F")
            else:
                # loop over all datasets
                for datasetSpec in taskSpec.datasetSpecList:
                    # only output and log datasets
                    if datasetSpec.type not in ["log", "output"]:
                        continue
                    tmpLog.info(f"freezing datasetID={datasetSpec.datasetID}:Name={datasetSpec.datasetName}")
                    ddmIF.freezeDataset(datasetSpec.datasetName, ignoreUnknown=True)
        except Exception as e:
            tmpLog.warning(f"failed to freeze datasets with {str(e)}")
            return self.SC_FAILED
        try:
            self.doBasicPostProcess(taskSpec, tmpLog)
        except Exception as e:
            tmpLog.error(f"doBasicPostProcess failed with {str(e)}")
            return self.SC_FATAL
        return self.SC_SUCCEEDED
